# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import concurrent
import datetime
import logging
import multiprocessing as mp
import os
import pickle
import re
from collections import deque
from concurrent.futures import ProcessPoolExecutor
from decimal import Decimal

from supplychainpy import model_inventory
from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_APPLICATION_CONFIG, ABS_FILE_PICKLE
from supplychainpy._helpers._pickle_config import deserialise_config
from supplychainpy.bi.recommendation_generator import run_sku_recommendation, run_profile_recommendation
from supplychainpy.inventory.analyse_uncertain_demand import UncertainDemand
from supplychainpy.inventory.summarise import Inventory
from supplychainpy.reporting.app import create_app
from supplychainpy.reporting.blueprints.models import Currency
from supplychainpy.reporting.blueprints.models import Forecast
from supplychainpy.reporting.blueprints.models import ForecastBreakdown
from supplychainpy.reporting.blueprints.models import ForecastStatistics
from supplychainpy.reporting.blueprints.models import ForecastType
from supplychainpy.reporting.blueprints.models import InventoryAnalysis
from supplychainpy.reporting.blueprints.models import MasterSkuList
from supplychainpy.reporting.blueprints.models import Orders
from supplychainpy.reporting.blueprints.models import TransactionLog, Recommendations, ProfileRecommendation
from supplychainpy.reporting.extensions import db
from supplychainpy.sample_data.config import ABS_FILE_PATH

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')


def currency_codes() -> dict:
    """ Retrives HTML Entity (decimal) for currency symbol.

    Returns:
        dict: Currency Symbols.

    """
    codes = {"AED": ("United Arab Emirates Dirham", "&#92;&#117;&#48;&#54;&#50;&#102;&#46;"),
             "ANG": ("Netherlands Antilles Guilder", "&#402"),
             "EUR": ("Euro Member Countries", "&#8364;"),
             "GBP": ("United Kingdom Pound", "&#163;"),
             "USD": ("United States Dollar", "&#36;"),
             }
    return codes


def _analysis_forecast_simple(analysis: UncertainDemand) -> dict:
    """ Retrieves simple_exponential_forecast from an instance of UncertainDemand.
        Function only required for Concurrent.futures.

    Args:
        analysis (UncertainDemand): Instance of UncertainDemand

    Returns:
        dict:   Forecast breakdown.
    """
    logging.log(logging.INFO,
                "Simple exponential smoothing forecast for SKU: {}\nObject id: {} ".format(
                    analysis.sku_id,
                    id(analysis)
                )
                )
    return analysis.simple_exponential_smoothing_forecast


def _analysis_forecast_holt(analysis: UncertainDemand) -> dict:
    """ Retrieves holts_exponential_forecast from an instance of UncertainDemand.
        Function only required for Concurrent.futures.

    Args:
        analysis (UncertainDemand): Instance of UncertainDemand

    Returns:
        dict:   Forecast breakdown.
    """
    logging.log(logging.INFO,
                "Holt's trend corrected exponential smoothing forecast for SKU: {}\nObject id: {} ".format(
                    analysis.sku_id,
                    id(analysis)
                )
                )
    return analysis.holts_trend_corrected_forecast


def load_currency(fx_codes: currency_codes(), ctx: db):
    """Loads Currency Symbols"""

    for key, value in fx_codes.items():
        codes = Currency()
        codes.country = value[0]
        codes.symbol = value[1]
        codes.currency_code = key
        ctx.session.add(codes)
    ctx.session.commit()


def batch(analysis, n):
    """Yield n-sized batches from analysis.
    Args:
        analysis:
        n:
    """

    for i in range(0, len(analysis), n):
        yield analysis[i:i + n]


def parallelise_ses(pickled_ses_batch_files: list, core_count: int) -> dict:
    """ Execute the exponential smoothing forecast in parallel.

    Args:
        pickled_ses_batch_files:   Uncertain demand objects batched for appropriate number of cores available on
                                    the host machine
        core_count:                 Number of cores available on the host machine, minus one.

    Returns:
        dict
    """
    attempts = 0
    simple_forecast = {}
    pickled_ses_batch_files_completed = []
    log.log(logging.INFO, 'Multiprocessing batch {} files using {} cores'.format(len(pickled_ses_batch_files), core_count))
    try:
        for num, batch_path in enumerate(pickled_ses_batch_files):
            order_batch = read_pickle(batch_path)[0]
            with ProcessPoolExecutor(max_workers=core_count) as executor:
                simple_forecast = {}
                ses_forecast_futures = {analysis.sku_id: executor.submit(_analysis_forecast_simple, analysis) for
                                        analysis in order_batch}
                ses_forecast_gen = {future: concurrent.futures.as_completed(ses_forecast_futures[future], timeout=20)
                                    for future
                                    in ses_forecast_futures}
                simple_forecast.update(
                    {value: ses_forecast_futures[value].result(timeout=30, ) for value in ses_forecast_gen})
                build_results_pickle(simple_forecast)
                log.log(logging.INFO, 'Batch processed: {}'.format(batch_path))
                pickled_ses_batch_files_completed.append(batch_path)
                # simple_forecast.clear()
                # ses_forecast_gen.clear()
                # ses_forecast_futures.clear()
                # del ses_forecast_futures
                # del ses_forecast_gen
                # del simple_forecast

        simple_forecast = retrieve_results_pickle()
        simple_forecast = simple_forecast[0]

        return simple_forecast
    except concurrent.futures.TimeoutError as err:
        print(err)
        attempts +=1
        pickled_ses_batch_files_remaining = set(pickled_ses_batch_files) -  set(pickled_ses_batch_files_completed)
        log.log(logging.INFO, 'ERROR processing batch. Retrying remaining batch files: {}'.format(pickled_ses_batch_files_remaining))
        for file_path in pickled_ses_batch_files_completed:
            remove_pickle(file_path)

        if len(pickled_ses_batch_files_remaining) > 0 and attempts > 4:
            parallelise_ses(pickled_ses_batch_files=list(pickled_ses_batch_files_remaining), core_count=core_count)
        else:
            print('The forecasting calculation process was unable to complete. Please check the source file.')
            return {}
    except OSError as err:
        print(err)


def parallelise_htc(batched_analysis: list, core_count: int):
    """ Execute the Holts' trend corrected smoothing forecast in parallel.

    Args:
        batched_analysis: Uncertain demand objects batched for appropriate number of cores available on the host machine
        core_count: Number of cores available on the host machine, minus one.

    Returns:
        dict
    """
    holts_forecast = {}
    try:
        for order_batch in batched_analysis:
            with ProcessPoolExecutor(max_workers=core_count) as executor:
                holts_forecast_futures = {analysis.sku_id: executor.submit(_analysis_forecast_holt, analysis) for
                                          analysis in order_batch}
                holts_forecast_gen = {future: concurrent.futures.as_completed(holts_forecast_futures[future]) for future
                                      in holts_forecast_futures}
                holts_forecast.update({value: holts_forecast_futures[value].result() for value in holts_forecast_gen})
    except TypeError as err:
        print('{}'.format(err))
    return holts_forecast


def write_pickle(**kwargs) -> str:
    """pickle data to file"""
    try:
        for k, v in kwargs.items():
            path = os.path.abspath(os.path.join(ABS_FILE_PICKLE, k))
            log.log(logging.INFO, 'Pickled file created at: {}'.format(path))
            with open(path, "wb") as ses:
                pickle.dump(v, ses)
            return path
    except OSError as err:
        print(err)
        return ''


def read_pickle(batch_path: str):
    """Read pickled data from file"""
    retrieved_pickle = []
    try:
        with open(batch_path, "r+b") as ses:
            log.log(logging.INFO, 'Pickled file retrieved from : {}'.format(ses))
            retrieved_pickle.append(pickle.load(ses))
        return retrieved_pickle
    except OSError as err:
        print(err)
        return retrieved_pickle


def remove_pickle(path: str):
    try:
        os.remove(path=path)
    except OSError as err:
        print('file not present. {}'.format(err))
        pass


def pickle_ses_forecast(batched_analysis: list) -> list:
    pickled_paths = []
    for num, item in enumerate(batched_analysis):
        filename = 'ses{}.pickle'.format(num)
        pickle_me = {filename: item}
        pickle_path = write_pickle(**pickle_me)
        pickled_paths.append(pickle_path)
    return pickled_paths


def build_results_pickle(ses_forecast_results: dict):
    filename = 'ses_forecast_results'
    path = os.path.abspath(os.path.join(ABS_FILE_PICKLE, filename))
    log.log(logging.INFO, 'Generating results pickle at: {}'.format(path))
    stored_ses_data = read_pickle(path)
    for i in stored_ses_data:
        ses_forecast_results.update(i)
    with open(path, "wb") as ses:
        pickle.dump(ses_forecast_results, ses)


def retrieve_results_pickle():
    filename = 'ses_forecast_results'
    path = os.path.abspath(os.path.join(ABS_FILE_PICKLE, filename))
    stored_ses_data = read_pickle(path)
    remove_pickle(path=path)
    return stored_ses_data


def cleanup_pickled_files():
    try:
        filename = 'ses_forecast_results'
        path = os.path.abspath(os.path.join(ABS_FILE_PICKLE, filename))
        remove_pickle(path=path)
        ses_file_regex = re.compile('[s][e][s]\d+[.][p][i][c][k][l][e]')
        for filename in os.listdir(ABS_FILE_PICKLE):
            if ses_file_regex.match(filename) and filename.endswith(".pickle"):
                path = os.path.abspath(os.path.join(ABS_FILE_PICKLE, filename))
                log.log(logging.INFO, 'Cleaning up pickled file at: {}'.format(path))
                remove_pickle(path)

    except FileNotFoundError as err:
        print(err)
        pass
    except OSError as err:
        print(err)


def load(file_path: str, location: str = None):
    """ Loads analysis and forecast into local database for reporting suite.

    Args:
        file_path (str):    File path to source file containing data for analysis.
        location (str):     Location of database to populate.

    """
    try:
        app = create_app()
        db.init_app(app)
        if location is not None and os.name in ['posix', 'mac']:
            app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///{}/reporting.db'.format(location)

        elif location is not None and os.name == 'nt':
            app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///{}\\reporting.db'.format(location)

        log.log(logging.DEBUG, 'Loading data analysis for reporting suite... \n')

        with app.app_context():
            db.create_all()
            log.log(logging.DEBUG, 'loading currency symbols...\n')
            print('loading currency symbols...\n')
            fx = currency_codes()
            load_currency(fx, db)

            print('loading currency symbols...[COMPLETED]\n')
            config = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
            currency = config.get('currency')

            log.log(logging.DEBUG, 'Analysing file: {}...\n'.format(file_path))
            print('Analysing file: {}...\n'.format(file_path))
            orders_analysis = model_inventory.analyse(file_path=file_path,
                                                      z_value=Decimal(1.28),
                                                      reorder_cost=Decimal(5000),
                                                      file_type="csv", length=12, currency=currency)

            ia = [analysis.orders_summary() for analysis in orders_analysis]
            date_now = datetime.datetime.now()
            analysis_summary = Inventory(processed_orders=orders_analysis)
            print('Analysing file: {}...[COMPLETED]\n\nCalculating Forecasts...\n'.format(file_path))
            log.log(logging.DEBUG, 'Calculating Forecasts...\n')

            cores = int(mp.cpu_count())
            cores -= 1
            batched_analysis = [i for i in batch(orders_analysis, cores)]
            pickled_paths = pickle_ses_forecast(batched_analysis=batched_analysis)
            try:
                simple_forecast = parallelise_ses(pickled_ses_batch_files=pickled_paths, core_count=cores)
            except:
                simple_forecast = {}
            cleanup_pickled_files()

            try:
                holts_forecast = parallelise_htc(batched_analysis=batched_analysis, core_count=cores)
            except:
                holts_forecast = {}

            transact = TransactionLog()
            transact.date = date_now
            db.session.add(transact)
            db.session.commit()

            transaction_sub = db.session.query(db.func.max(TransactionLog.date))
            transaction_id = db.session.query(TransactionLog).filter(TransactionLog.date == transaction_sub).first()
            load_profile_recommendations(analysed_order=orders_analysis, forecast=holts_forecast,
                                         transaction_log_id=transaction_id)

            # d = _Orchestrate()
            # d.update_database(int(transaction_id.id))
            forecast_types = ('ses', 'htces')
            for f_type in forecast_types:
                forecast_type = ForecastType()
                forecast_type.type = f_type
                db.session.add(forecast_type)
            db.session.commit()
            ses_id = db.session.query(ForecastType.id).filter(ForecastType.type == forecast_types[0]).first()
            htces_id = db.session.query(ForecastType.id).filter(ForecastType.type == forecast_types[1]).first()
            current_msk = db.session.query(MasterSkuList.sku_id).all()
            match_sku = [str(i) for i in current_msk]
            print(match_sku)
            print('Calculating Forecasts...[COMPLETED]\n')
            log.log(logging.DEBUG, 'loading database ...\n')
            print('loading database ...\n')

            for item in ia:
                skus_description = [summarised for summarised in analysis_summary.describe_sku(item['sku'])]
                denom = db.session.query(Currency.id).filter(Currency.currency_code == item['currency']).first()
                master_sku = MasterSkuList()
                master_sku.sku_id = item['sku']
                db.session.add(master_sku)
                i_up = InventoryAnalysis()
                mk = db.session.query(MasterSkuList.id).filter(MasterSkuList.sku_id == item['sku']).first()
                i_up.sku_id = mk.id
                tuple_orders = item['orders']
                # print(tuple_orders)
                i_up.abc_xyz_classification = item['ABC_XYZ_Classification']
                i_up.standard_deviation = item['standard_deviation']
                i_up.backlog = item['backlog']
                i_up.safety_stock = item['safety_stock']
                i_up.reorder_level = item['reorder_level']
                i_up.economic_order_quantity = item['economic_order_quantity']
                i_up.demand_variability = item['demand_variability']
                i_up.average_orders = round(float(item['average_orders']))
                i_up.shortages = item['shortages']
                i_up.excess_stock = item['excess_stock']
                i_up.reorder_quantity = item['reorder_quantity']
                i_up.economic_order_variable_cost = item['economic_order_variable_cost']
                i_up.unit_cost = item['unit_cost']
                i_up.revenue = item['revenue']
                i_up.date = date_now
                i_up.safety_stock_rank = skus_description[0]['safety_stock_rank']
                i_up.shortage_rank = skus_description[0]['shortage_rank']
                i_up.excess_cost = skus_description[0]['excess_cost']
                i_up.percentage_contribution_revenue = skus_description[0]['percentage_contribution_revenue']
                i_up.excess_rank = skus_description[0]['excess_rank']
                i_up.retail_price = skus_description[0]['retail_price']
                i_up.gross_profit_margin = skus_description[0]['gross_profit_margin']
                i_up.min_order = skus_description[0]['min_order']
                i_up.safety_stock_cost = skus_description[0]['safety_stock_cost']
                i_up.revenue_rank = skus_description[0]['revenue_rank']
                i_up.markup_percentage = skus_description[0]['markup_percentage']
                i_up.max_order = skus_description[0]['max_order']
                i_up.shortage_cost = skus_description[0]['shortage_cost']
                i_up.quantity_on_hand = item['quantity_on_hand']
                i_up.currency_id = denom.id
                i_up.traffic_light = skus_description[0]['inventory_traffic_light']
                i_up.inventory_turns = skus_description[0]['inventory_turns']
                i_up.transaction_log_id = transaction_id.id
                db.session.add(i_up)
                inva = db.session.query(InventoryAnalysis.id).filter(InventoryAnalysis.sku_id == mk.id).first()
                for i, t in enumerate(tuple_orders['demand'], 1):
                    orders_data = Orders()
                    # print(r)
                    orders_data.order_quantity = t
                    orders_data.rank = i
                    orders_data.analysis_id = inva.id
                    db.session.add(orders_data)
                    # need to select sku id
                if simple_forecast is not None:
                    for i, forecasted_demand in enumerate(simple_forecast, 1):
                        if forecasted_demand == item['sku']:
                            forecast_stats = ForecastStatistics()
                            forecast_stats.analysis_id = inva.id
                            forecast_stats.mape = simple_forecast.get(forecasted_demand)['mape']
                            forecast_stats.forecast_type_id = ses_id.id
                            forecast_stats.slope = simple_forecast.get(forecasted_demand)['statistics']['slope']
                            forecast_stats.p_value = simple_forecast.get(forecasted_demand)['statistics']['pvalue']
                            forecast_stats.test_statistic = simple_forecast.get(forecasted_demand)['statistics'][
                                'test_statistic']
                            forecast_stats.slope_standard_error = simple_forecast.get(forecasted_demand)['statistics'][
                                'slope_standard_error']
                            forecast_stats.intercept = simple_forecast.get(forecasted_demand)['statistics']['intercept']
                            forecast_stats.standard_residuals = simple_forecast.get(forecasted_demand)['statistics'][
                                'std_residuals']
                            forecast_stats.trending = simple_forecast.get(forecasted_demand)['statistics']['trend']
                            forecast_stats.optimal_alpha = simple_forecast.get(forecasted_demand)['optimal_alpha']
                            forecast_stats.optimal_gamma = 0
                            db.session.add(forecast_stats)
                            for p in range(0, len(simple_forecast.get(forecasted_demand)['forecast'])):
                                forecast_data = Forecast()
                                forecast_data.forecast_quantity = simple_forecast.get(forecasted_demand)['forecast'][p]
                                forecast_data.analysis_id = inva.id
                                forecast_data.forecast_type_id = ses_id.id
                                forecast_data.period = p + 1
                                forecast_data.create_date = date_now
                                db.session.add(forecast_data)
                            for q, sesf in enumerate(simple_forecast.get(forecasted_demand)['forecast_breakdown']):
                                forecast_breakdown = ForecastBreakdown()
                                forecast_breakdown.analysis_id = inva.id
                                forecast_breakdown.forecast_type_id = ses_id.id
                                forecast_breakdown.trend = 0
                                forecast_breakdown.period = sesf['t']
                                forecast_breakdown.level_estimates = \
                                    sesf['level_estimates']
                                forecast_breakdown.one_step_forecast = \
                                    sesf['one_step_forecast']
                                forecast_breakdown.forecast_error = \
                                    sesf['forecast_error']
                                forecast_breakdown.squared_error = sesf['squared_error']
                                forecast_breakdown.regression = simple_forecast.get(forecasted_demand)['regression'][q]
                                db.session.add(forecast_breakdown)
                            break

                for i, holts_forecast_demand in enumerate(holts_forecast, 1):
                    if holts_forecast_demand == item['sku']:
                        forecast_stats = ForecastStatistics()
                        forecast_stats.analysis_id = inva.id
                        forecast_stats.mape = holts_forecast.get(holts_forecast_demand)['mape']
                        forecast_stats.forecast_type_id = htces_id.id
                        forecast_stats.slope = holts_forecast.get(holts_forecast_demand)['statistics'].get('slope')
                        forecast_stats.p_value = holts_forecast.get(holts_forecast_demand)['statistics'].get('pvalue')
                        forecast_stats.test_statistic = holts_forecast.get(holts_forecast_demand)['statistics'].get(
                            'test_statistic')
                        forecast_stats.slope_standard_error = holts_forecast.get(holts_forecast_demand)['statistics'].get(
                            'slope_standard_error')
                        forecast_stats.intercept = holts_forecast.get(holts_forecast_demand)['statistics'].get('intercept')
                        forecast_stats.standard_residuals = holts_forecast.get(holts_forecast_demand)['statistics'].get('std_residuals')
                        forecast_stats.trending = holts_forecast.get(holts_forecast_demand)['statistics'].get('trend')
                        forecast_stats.optimal_alpha = holts_forecast.get(holts_forecast_demand)['optimal_alpha']
                        forecast_stats.optimal_gamma = holts_forecast.get(holts_forecast_demand)['optimal_gamma']
                        db.session.add(forecast_stats)
                        for p in range(0, len(holts_forecast.get(holts_forecast_demand)['forecast'])):
                            forecast_data = Forecast()
                            forecast_data.forecast_quantity = holts_forecast.get(holts_forecast_demand)['forecast'][p]
                            forecast_data.analysis_id = inva.id
                            forecast_data.forecast_type_id = htces_id.id
                            forecast_data.period = p + 1
                            forecast_data.create_date = date_now
                            db.session.add(forecast_data)
                        for k, htcesf in enumerate(holts_forecast.get(holts_forecast_demand)['forecast_breakdown']):
                            forecast_breakdown = ForecastBreakdown()
                            forecast_breakdown.analysis_id = inva.id
                            forecast_breakdown.forecast_type_id = htces_id.id
                            forecast_breakdown.trend = htcesf['trend']
                            forecast_breakdown.period = htcesf['t']
                            forecast_breakdown.level_estimates = \
                                htcesf['level_estimates']
                            forecast_breakdown.one_step_forecast = \
                                htcesf['one_step_forecast']
                            forecast_breakdown.forecast_error = \
                                htcesf['forecast_error']
                            forecast_breakdown.squared_error = htcesf['squared_error']
                            forecast_breakdown.regression = holts_forecast.get(holts_forecast_demand)['regression'][k]
                            db.session.add(forecast_breakdown)
                        break

            db.session.commit()
            print('loading database ...[COMPLETED]\n')
            loading = 'Loading recommendations into database...'
            print(loading, end="")
            load_recommendations(summary=ia, forecast=holts_forecast, analysed_order=orders_analysis)
            print('Loading recommendations into database...[COMPLETED]\n')
            log.log(logging.DEBUG, 'Analysis ...\n')
            print('Analysis ... [COMPLETED]\n')
    except OSError as e:
        print(e)


def load_recommendations(summary, forecast, analysed_order):
    recommend = run_sku_recommendation(analysed_orders=analysed_order, forecast=forecast)
    for item in summary:
        rec = Recommendations()
        mk = db.session.query(MasterSkuList.id).filter(MasterSkuList.sku_id == item['sku']).first()
        inva = db.session.query(InventoryAnalysis.id).filter(InventoryAnalysis.sku_id == mk.id).first()
        rec.analysis_id = inva.id
        reco = 'There are no recommendation at this time.' if recommend.get(item['sku'],
                                                                            'There are no recommendation at this time.'
                                                                            ) == '' else recommend.get(item['sku'],
                                                                                                       'None')
        rec.statement = reco
        db.session.add(rec)
        db.session.commit()


def load_profile_recommendations(analysed_order, forecast, transaction_log_id):
    recommend = run_profile_recommendation(analysed_orders=analysed_order, forecast=forecast)
    rec = ProfileRecommendation()
    rec.transaction_id = int(transaction_log_id.id)
    rec.statement = recommend.get('profile')
    db.session.add(rec)
    db.session.commit()


if __name__ == '__main__':
    load(ABS_FILE_PATH['COMPLETE_CSV_SM'])



import flask
from flask import Blueprint


contact_blueprint = Blueprint('contact', __name__, template_folder='templates')


@contact_blueprint.route('/about', methods=['GET'])
def about():
    return flask.render_template('contact/about.html')



import multiprocessing as mp


class CoreCount:
    """ Manage cpu core count for setting multiprocessing tasks.

    """
    def __init__(self):
        self._core_count = self._get_core_count()

    @property
    def core_count(self) ->int:
        """Returns system core count.
        """
        return self._core_count

    @staticmethod
    def _get_core_count()->int:
        return int(mp.cpu_count())

if __name__ =="__main__":
    c = CoreCount()
    print(c.core_count)


# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from copy import deepcopy

import logging
import numpy as np

from supplychainpy._helpers._decorators import log_this

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

class _PairwiseComparison:
    __UNKNOWN = "unknown"
    __RANDOM_INDICES = {1:0, 2: 0, 3: 0.58, 4: 0.90, 5: 1.12, 6: 1.24, 7: 1.32, 8: 1.41, 9: 1.45, 10: 1.51}

    def __init__(self, criteria: tuple, criteria_scores: list, options: tuple, option_scores: dict,
                 quantitative_criteria: tuple = None):
        self._criteria = criteria
        self._importance = criteria_scores
        self._alternatives = options
        self._alternative_scores = option_scores
        self._quantitative_criteria = quantitative_criteria
        self._cr = self._consistency_ratio()

    @property
    def criteria(self) -> tuple:
        return self._criteria

    @criteria.setter
    def criteria(self, criteria: tuple):
        self._criteria = criteria

    @property
    def importance(self) -> list:
        return self._importance

    @importance.setter
    def importance(self, importance: list):
        self._importance = importance

    @property
    def alternatives(self) -> tuple:
        return self._alternatives

    @alternatives.setter
    def alternatives(self, alternatives: tuple):
        self._alternatives = alternatives

    @property
    def alternative_scores(self) -> dict:
        return self._alternative_scores

    @alternative_scores.setter
    def alternative_scores(self, alternative_scores: dict):
        self._alternative_scores = alternative_scores

    @property
    def quantitative_criteria(self) -> tuple:
        return self._quantitative_criteria

    @quantitative_criteria.setter
    def quantitative_criteria(self, quantitative_criteria: tuple):
        self.quantitative_criteria = quantitative_criteria

    @property
    def consistency_ratio(self):
        return self._cr

    @log_this(logging.DEBUG, message='Called to map the reciprocal in pairwise matrix.')
    def _map_reciprocal(self, importance_score: list) -> np.array:
        """ Fills in the reciprocal relative weights matrix of  for each criterion.

        Args:
            importance_score:   The relative scores for each criterion.

        Returns:
            np.array:           A matrix of scores indicating the relative importance of the criterion.
        """
        log.log(logging.INFO, "mapping reciprocals to pairwise matrix")
        importance_score_matrix = np.array(importance_score)
        for index_one, score_one in enumerate(importance_score):
            for index_two, score_two in enumerate(importance_score):
                if index_two >= index_one:
                    importance_score_matrix[index_two, index_one] = self._reciprocal(
                        importance_score_matrix[index_one, index_two])
        return importance_score_matrix

    @log_this(logging.DEBUG, message='Called to compile the solution with scores for the relative importance '
                                    'for each category.')
    def summary(self) -> dict:
        """ Summarise the Analytical Hierarchy Process ,with a dict of scores for ranking.

        Returns:
            dict:   Scores for each alternative based on the relative importance of each category.
        """
        log.log(logging.INFO, "Scoring alternative for ranking")
        np.set_printoptions(precision=3, suppress=True)
        criteria_eigenvector_rank = self.compute_criteria_eingenvector()
        alternative_matrix_subjective = {val: self._square_matrix(np.array(self.alternative_scores.get(val))) for val in
                                         self.alternative_scores.keys() if val not in self.quantitative_criteria}

        alternative_matrix_eigenvectors = self._alternative_eigenvector(alternative_matrix_subjective)
        alternative_matrix_quantitative = self._normalise_quantitative_rank()

        recompile_main_hierarchy_matrix = self._recompile_main_hierarchy_matrix(alternative_matrix_eigenvectors,
                                                                                alternative_matrix_quantitative)

        ahp_solution = self._compile_ahp_solution(recompiled_main_hierarchy=recompile_main_hierarchy_matrix,
                                                  criteria_eigenvector=criteria_eigenvector_rank)
        return ahp_solution

    @staticmethod
    @log_this(logging.DEBUG, "Called to compute cost benefit ratio for alternative in AHP.")
    def cost_benefit_summary(ahp_summary: dict, item_cost: dict) -> dict:
        """ Normalises the costs for each alternative and converts cost into benefits cost ratio.

        Args:
            ahp_summary (dict):     Summary of analytical hierarchy.
            item_cost (dict):       Cost of each alternative

        Returns:
            dict:   benefit cost ratio for alternatives in analytical hierarchy

        """
        log.log(logging.INFO, "Cost benefits analysis")
        if ahp_summary.keys() == item_cost.keys():
            total_cost = sum([item_cost.get(i) for i in item_cost])
            normalised_item_cost = {}
            for i in item_cost:
                normalised_item_cost.update({i: item_cost.get(i) / total_cost})
            cost_benefit_ratios = {}
            for i in normalised_item_cost:
                cost_benefit_ratios.update({i: ahp_summary.get(i) / normalised_item_cost.get(i)})
            return cost_benefit_ratios
        else:
            raise KeyError("Please check keys and ensure both 'ahp_summary' and 'item_costs' have identical keys.")

    @log_this(logging.DEBUG, "Called to compile overall solution,")
    def _compile_ahp_solution(self, recompiled_main_hierarchy: dict, criteria_eigenvector: tuple) -> dict:
        """ Multiplies the criteria rankings by the rankings for the alternatives.

        Args:
            recompiled_main_hierarchy (dict):   The eigenvectors for the subjective and objective categories
            criteria_eigenvector (dict):        The eigenvector for the criteria.

        Returns:
            dict:   The final solution for the AHP.

        """
        log.log(logging.INFO, "Compile solution as dict")
        ahp_solution = {}
        stack = []
        for count, alternative in enumerate(self.alternatives):
            for index, name in enumerate(self.criteria):
                stack.append(recompiled_main_hierarchy.get(name)[count] * criteria_eigenvector[index])
            ahp_solution.update({alternative: sum(stack)})
            stack.clear()
        return ahp_solution

    @log_this(logging.DEBUG, "Called to compute cost benifit ratio for alternatve in AHP.")
    def _alternative_eigenvector(self, alternative_matrix):
        """

        Args:
            alternative_matrix:

        Returns:

        """
        np.set_printoptions(precision=3, suppress=True)
        eigenvectors_for_alternatives = []
        for x in alternative_matrix.keys():
            eigenvectors_for_alternatives.append({x: self._calculate_eigenvector(alternative_matrix.get(x))})
        return eigenvectors_for_alternatives

    @log_this(logging.DEBUG, "Called to compute normalised ranking.")
    def _normalise_quantitative_rank(self) -> list:
        """ Normalises the values for the categories ranked using quantities e.g. fuel consumption.

        Returns:
            list:   Normalised ranking of alternatives for quantitative categories.

        """
        log.log(logging.INFO, "Compile normalised ranking")
        normalised_quantitive_criteria_ranking = []
        for x in self.quantitative_criteria:
            quant_criteria = self.alternative_scores.get(x)
            sum_quant_criteria = sum(quant_criteria)
            normalised_quantitive_criteria = tuple([i / sum_quant_criteria for i in quant_criteria])
            normalised_quantitive_criteria_ranking.append({x: normalised_quantitive_criteria})
        return normalised_quantitive_criteria_ranking

    @staticmethod
    @log_this(logging.DEBUG, "Called to combine subjective and objective criteria matrices.")
    def _recompile_main_hierarchy_matrix(alternative_matrix_quantitative: list,
                                         alternative_matrix_eigenvectors: list) -> dict:
        """ Combines the subjective matrices and the objective (quantitative) matrices together.

        Args:
            alternative_matrix_quantitative (list):     Eigenvectors for the objective categories for the AHP
            alternative_matrix_eigenvectors (list):     Eigenvectors for the subjective categories for the AHP

        Returns:
            dict:   Combined subjective and objective eigenvectors.

        """
        log.log(logging.INFO, "Combine the subjective matrices")
        for i in alternative_matrix_quantitative:
            alternative_matrix_eigenvectors.append(i)
        recompiled_matrix = {}
        for x in alternative_matrix_eigenvectors:
            recompiled_matrix.update(x)
        return recompiled_matrix

    @staticmethod
    def _reciprocal(first_score: float) -> float:
        return 1 / first_score

    @staticmethod
    def _square_matrix(score: np.array) -> np.array:
        """ Matrix multiplication

        Args:
            score: completed scores for comparison matrix

        Returns:
            np.array:   squared completed scores for comparison matrix

        """
        return score * score

    @log_this(logging.DEBUG, "Called to calculate eigenvector.")
    def _calculate_eigenvector(self, squared_comparison_matrix: np.array) -> tuple:
        """ Calculates the eigenvector.

        Args:
            squared_comparison_matrix (np.array):   Squared comparison matrix.

        Returns:
            tuple:  Eigenvector.

        """
        log.log(logging.INFO, "calculating eigenvector")
        sum_block = [sum(i) for i in squared_comparison_matrix]
        block_total = sum(sum_block)
        eigenvector = [i / block_total for i in sum_block]
        last_result = eigenvector

        while True:

            new_matrix = self._square_matrix(squared_comparison_matrix)
            sum_block = [sum(i) for i in new_matrix]
            block_total = sum(sum_block)
            new_eigenvector = [i / block_total for i in sum_block]
            if new_eigenvector == last_result:
                break
            else:
                last_result = deepcopy(new_eigenvector)
                new_eigenvector.clear()
        # comparison_eigenvector = namedtuple('comparison_eigenvector', [*self.criteria])
        # final_eigenvector = comparison_eigenvector(*eigenvector)
        return tuple(eigenvector)

    @log_this(logging.DEBUG, "Called to initiate AHP.")
    def compute_criteria_eingenvector(self) -> tuple:
        """ Initiates and conducts the calculation of eigenvectors.

        Returns:
            tuple: Eigenvector.

        """
        log.log(logging.INFO, "Starting AHP")
        if self._criteria != self.__UNKNOWN and self._importance != self.__UNKNOWN:
            if isinstance(self._criteria, tuple):
                comparison = self.pairwise_tp(score=self._importance)
                squared_comparison_matrix = self._square_matrix(score=comparison)
                eigenvector_ranking = self._calculate_eigenvector(squared_comparison_matrix=squared_comparison_matrix)
                return eigenvector_ranking
            else:
                raise TypeError("The criteria supplied as an argument is an incorrect type. "
                                "Please supply a correct type.")
        else:
            raise KeyError("Please supply criteria and importance.")

    def pairwise_tp(self, score: list) -> tuple:
        """ Creates the comparison matrix by filling in the reciprocal values.

        Args:

            score (list):   Scores for criteria

        Returns:
            tuple:  completed comparisons.

        """
        comparison_matrix = self._map_reciprocal(importance_score=score)
        return comparison_matrix

    @log_this(logging.DEBUG, "Checks consistency of the choices made in the pair-wise matrix.")
    def _consistency_ratio(self):
        """ Calculates the consistency ratio, indicating how consistent the decision maker is being with their pair-wise
        comparisons.

        Returns:
            float:  Consistency ratio indicating the consistency of the pair-wise comparison.

        """
        comparison = self.pairwise_tp(score=self._importance)
        squared_comparison_matrix = self._square_matrix(score=comparison)
        eigenvector_ranking = self._calculate_eigenvector(squared_comparison_matrix=squared_comparison_matrix)
        sum_comparison = np.sum(squared_comparison_matrix, axis=0)
        lambda_max = sum(sum_comparison * np.array(eigenvector_ranking))
        consistency_index = (lambda_max - len(self.criteria)) / (len(self.criteria) - 1)
        consistency_ratio = consistency_index / self.__RANDOM_INDICES.get(len(self.criteria))
        return consistency_ratio




# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import uuid

from supplychainpy._helpers._pickle_config import deserialise_config, serialise_config
from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_APPLICATION_CONFIG, ABS_FILE_PATH_ARCHIVE
from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_CSV_MANAGEMENT_CONFIG
from sqlalchemy import create_engine

# Centralises all the construction of connections to the various database bases.

case = {

    'UNKNOWN': 'unknown',
    'REPORTING': 'reporting',
    'MANAGEMENT': 'management',
    'POSIX': 'posix',
    'MACINTOSH': 'mac',
    'NT': 'nt',
    'RETRIEVAL': 'retrieval',
    'CREATE': 'create'
}


def database_connection_uri(db: str = 'reporting', **kwargs) -> str:
    """ Manages the construction of a database uri for the purposes of connecting to a database.

    Args:
        db:     The string indicating 'reporting' or 'management' database uri.

    Returns:
        str:    database uri.

    """
    retrieve = kwargs.get('retrieve', False)
    create = kwargs.get('create', False)

    db_uri = ''

    unix, windows = [case['POSIX'], case['MACINTOSH']], [case['NT']]

    if retrieve:
        if db == case['REPORTING']:
            if os.name in unix:
                db_uri = _os_specific_uri(action_case=case['RETRIEVAL'], my_os=unix, type_case=case['REPORTING'])
            elif os.name in windows:
                db_uri = _os_specific_uri(action_case=case['RETRIEVAL'], my_os=windows, type_case=case['REPORTING'])
            return db_uri
        elif db == case['MANAGEMENT']:
            if os.name in unix:
                db_uri = _os_specific_uri(action_case=case['RETRIEVAL'], my_os=unix, type_case=case['MANAGEMENT'])
            elif os.name in windows:
                db_uri = _os_specific_uri(action_case=case['RETRIEVAL'], my_os=windows, type_case=case['MANAGEMENT'])
            return db_uri
        else:
            raise Exception("Incorrect database specified. Please select 'reporting' or 'management' ")
    elif create:
        if db == case['MANAGEMENT']:
            if os.name in unix:
                db_uri = _os_specific_uri(action_case=case['CREATE'], my_os=unix, type_case=case['MANAGEMENT'])
            elif os.name in windows:
                db_uri = _os_specific_uri(action_case=case['CREATE'], my_os=windows, type_case=case['MANAGEMENT'])
            return db_uri
    else:
        pass


# TODO-fix decide how the my_os variable should simplify this function, splitting the logic has increased lines of code.
def _os_specific_uri(action_case: str, my_os: list, type_case: str) -> str:
    """ Constructs os specific uri.
    Args:
        action_case:    The action base indicates wheter the connection is for retrieving from or creating a database
        my_os:          The os the user is running the library on.
        type_case:      Indicates wether the uri is for the csv management database or the reporting database

    Returns:

    """

    app_config = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
    man_config = deserialise_config(ABS_FILE_PATH_CSV_MANAGEMENT_CONFIG)
    db_uri = ''
    unix, windows = [case['POSIX'], case['MACINTOSH']], [case['NT']]

    if action_case == case['RETRIEVAL'] and type_case == case['REPORTING']:
        if os.name in unix:
            db_uri = 'sqlite:///{}/reporting.db'.format(app_config['database_path'])
        elif os.name in windows:
            db_uri = 'sqlite:///{}\\reporting.db'.format(app_config['database_path'])
        return db_uri
    elif action_case == case['RETRIEVAL'] and type_case == case['MANAGEMENT']:
        if os.name in unix:
            db_uri = 'sqlite:///{}/{}'.format(man_config['database_path'], man_config['database_name'])
        elif os.name in windows:
            db_uri = 'sqlite:///{}\\{}'.format(man_config['database_path'], man_config['database_name'])
        return db_uri
    elif action_case == case['CREATE'] and type_case == case['MANAGEMENT']:
        database_name = 'csv_management-{}.db'.format(uuid.uuid1())
        if os.name in unix:
            db_uri = 'sqlite:///{}/{}'.format(ABS_FILE_PATH_ARCHIVE, database_name)
        elif os.name in windows:
            db_uri = 'sqlite:///{}\\{}'.format(ABS_FILE_PATH_ARCHIVE, database_name)

        config = {
            'database_path': ABS_FILE_PATH_ARCHIVE,
            'database_name': database_name,
            'database_connection': db_uri
        }
        serialise_config(configuration=config, file_path=ABS_FILE_PATH_CSV_MANAGEMENT_CONFIG)

        return db_uri


def engine(uri):
    cr_engine = create_engine(uri)
    return cr_engine


if __name__ == '__main__':
    print(database_connection_uri('management', retrieve='retrieve'))



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from supplychainpy.inventory.units import SKU, Package


class Pallet:
    def __init__(self):
        self.length = None
        self.width = None
        self.height = None
        self.weight = None
        self.package = []

    def recalculate_weight(self):
        pass

    def recalculate_height(self):
        pass


class PalletBuilder:
    def __init__(self):
        self.pallet = None

    def make_new_pallet(self):
        self.pallet = Pallet()


class PalletDirector:
    def __init__(self, pallet_builder: PalletBuilder):
        self._pallet_builder = pallet_builder

    def construct_pallet(self, length: float, width: float, height: float, weight: float):
        self._pallet_builder.make_new_pallet()
        self._pallet_builder.pallet.height = height
        self._pallet_builder.pallet.length = length
        self._pallet_builder.pallet.width = width
        self._pallet_builder.pallet.width = weight

    def add_package(self, package: Package):
        self._pallet_builder.pallet.package.append(package)

    def recalculate_dimensions(self):
        self._pallet_builder.pallet.recalculate_weight()
        self._pallet_builder.pallet.recalculate_height()

    def __add__(self, other):
        self._pallet_builder.pallet.package.append(other)

    @property
    def pallet(self)-> Pallet():
        return self._pallet_builder.pallet


class ShippingContainer(Pallet):
    def __init__(self):
        super().__init__()

        # how to represent a unit of packaging
        # parameter for unt nase shape circular or square to calculate fill



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from math import sqrt

# return a tuple for variable cost and order quantity without sending
def minimum_variable_cost(double total_orders, double  reorder_cost, double unit_cost, double  holding_cost):

    cdef double STEP
    STEP = 0.2
    cdef double previous_eoq_variable_cost

    cdef double vc
    cdef int counter = 0
    cdef double order_qty

    previous_eoq_variable_cost = 0.0
    vc = 0.0
    while previous_eoq_variable_cost >= vc:

        previous_eoq_variable_cost = vc
        # reorder cost * average demand all divided by order size + (demand size * holding cost)
        if counter < 1:
            order_qty = order_size(total_orders, reorder_cost, unit_cost, holding_cost)

        vc = variable_cost(total_orders, reorder_cost, order_qty, unit_cost, holding_cost)

        order_qty += order_qty * STEP

        if counter < 1:
            previous_eoq_variable_cost = vc

        while counter == 0:
            counter += 1

    return previous_eoq_variable_cost


def economic_order_quantity(double total_orders, double reorder_cost, double unit_cost, double holding_cost,
                            int reorder_quantity):

        cdef double STEP
        STEP= 0.2

        cdef double previous_eoq_variable_cost
        cdef double order_factor
        cdef double eoq_variable_cost

        cdef int counter = 0
        cdef double order_qty

        eoq_variable_cost = 0.0
        previous_eoq_variable_cost = 0.0

        while previous_eoq_variable_cost >= vc:

            previous_eoq_variable_cost = vc

            if counter < 1:
                order_qty = order_size(total_orders, reorder_cost, unit_cost, holding_cost)
                # print('1st order quantity {}'.format(order_qty))
            vc = variable_cost(total_orders, reorder_cost, order_qty, unit_cost, holding_cost)

            if counter >=1 :
                order_qty += order_qty * STEP

            if counter < 1:
                previous_eoq_variable_cost = vc

            while counter == 0:
                counter += 1

        return order_qty

cdef double variable_cost(double total_orders, double reorder_cost, double order_size, double  unit_cost,
                          double holding_cost):

    cdef double rc
    cdef double hc
    cdef double tvc

    rc = (total_orders * reorder_cost) / order_size
    hc = order_size * unit_cost * holding_cost
    tvc = rc + hc
    # print('reorder cost {}'.format(rc))
    # print('holding cost {}'.format(hc))
    # print('total variable cost {}'.format(tvc))
    return tvc

cdef double order_size(double total_orders, double  reorder_cost, double  unit_cost, double  holding_cost):

    cdef double order_qty

    order_qty = sqrt(((total_orders * reorder_cost * 2.0) / (unit_cost * holding_cost) ))* 0.4
    # print('order qty {}'.format(unit_cost * holding_cost))

    return order_qty




# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from operator import itemgetter

import itertools

def closing_stockout_percentage( list closing_stock, int period_length):

    cdef float closing_stock_count
    cdef float percentage

    closing_stock_count = closing_stock.count(0)
    percentage = closing_stock_count / period_length
    return percentage

def average_items(shortage_cost, int period_length):
    cdef float average
    average = sum(shortage_cost) / period_length
    return average


def optimum_std(float mean, list items):
    cdef list diff =[]
    cdef double variance, std, tmp
    cdef dict variance_analysis ={}

    for item in items:
        tmp = (item - mean)**2
        diff.append(tmp)

    variance = sum(diff)/len(diff)
    std = variance ** 0.5
    variance_analysis = {'variance': variance, 'standard_deviation': std}

    return variance_analysis


def summarize_monte_carlo(list simulation_frame, int period_length):
    """ summarises the monte carlo transaction summary

    Generates random distribution for demand over period specified and creates a simulation window for opening_stock,
    demand, closing_stock, delivery and backlog for each sku in the data set. Creates a transaction summary window
    for inventory movements.

    Args:
        simulation_frame (list):    The path to the file containing two columns of data, 1 period and 1 data-point per sku.
        period_length (int):        The number of periods define the simulation window e.g. 12 weeks, months etc.


    Returns:
        list:       A list containing the transaction summary for each period.

    """

    cdef:
        list closing_stock = [], opening_stock =[], shortage_units = [], summary = []
        list summarize = [], quantity_sold =[], backlog =[]

    cdef unsigned int i , n, x

    cdef:
        float cls, avg_ops, min_ops, max_ops, avg_backlog, min_backlog, max_backlog
        float avg_cls, min_cls, max_cls
        double shc, min_shc, max_shc, rev, min_quantity_sold, max_quantity_sold, var_ops, avg_shc, total_quantity_sold

    cdef dict std_ops ={}, std_backlog={}, std_cls={}, std_shc={}, std_quantity_sold={}

    n = int(len(simulation_frame[0]))//int(period_length) + 1
    i = 1
    for s in simulation_frame:
        for x in range(i, n):
            for f in s:
                if int(f[0]['index']) == x:
                    closing_stock.append( int(f[0]['closing_stock']))
                    shortage_units.append(float(f[0]['shortage_units']))
                    quantity_sold.append(int(f[0]['quantity_sold']))
                    opening_stock.append(int(f[0]['opening_stock']))
                    backlog.append(int(f[0]['backlog']))

                if len(closing_stock) == period_length and len(shortage_units) == period_length:
                    cls = closing_stockout_percentage(closing_stock, period_length)
                    avg_ops = average_items(opening_stock, period_length)
                    min_ops = min(opening_stock)
                    max_ops = max(opening_stock)
                    std_ops = optimum_std(avg_ops, opening_stock)
                    avg_backlog = average_items(backlog, period_length)
                    min_backlog = min(backlog)
                    max_backlog = max(backlog)
                    std_backlog = optimum_std(avg_backlog, backlog)
                    avg_cls = average_items(closing_stock, period_length)
                    min_cls = min(closing_stock)
                    max_cls = max(closing_stock)
                    std_cls = optimum_std(avg_cls, closing_stock)
                    shc = sum(shortage_units)
                    min_shc = min(shortage_units)
                    max_shc = max(shortage_units)
                    avg_shc = average_items(shortage_units, period_length)
                    std_shc = optimum_std(avg_shc, shortage_units)
                    total_quantity_sold = sum(quantity_sold)
                    avg_qty_sold = average_items(quantity_sold, period_length)
                    min_quantity_sold = min(quantity_sold)
                    max_quantity_sold = max(quantity_sold)
                    std_quantity_sold = optimum_std(avg_qty_sold, quantity_sold)

                    summary.append({'sku_id': f[0]['sku_id'],
                                    'standard_deviation_opening_stock': std_ops['standard_deviation'],
                                    'variance_opening_stock': std_ops['variance'],
                                    'stockout_percentage': cls,
                                    'average_closing_stock': avg_cls,
                                    'minimum_closing_stock': min_cls,
                                    'maximum_closing_stock': max_cls ,
                                    'standard_deviation_closing_stock': std_cls['standard_deviation'],
                                    'variance_closing_stock': std_cls['variance'],
                                    'total_shortage_units': shc,
                                    'standard_deviation_shortage_cost': std_shc['standard_deviation'],
                                    'variance_shortage_units': std_shc['variance'],
                                    'standard_deviation_revenue': std_quantity_sold['standard_deviation'],
                                    'variance_quantity_sold': std_quantity_sold['variance'],
                                    'minimum_shortage_units':min_shc,
                                    'maximum_shortage_units': max_shc,
                                    'average_opening_stock': avg_ops,
                                    'minimum_opening_stock': min_ops,
                                    'maximum_opening_stock': max_ops,
                                    'average_quantity_sold': avg_qty_sold,
                                    'minimum_quantity_sold': min_quantity_sold,
                                    'maximum_quantity_sold': max_quantity_sold,
                                    'average_backlog': avg_backlog,
                                    'minimum_backlog': min_backlog,
                                    'maximum_backlog': max_backlog,
                                    'standard_deviation_backlog': std_backlog['standard_deviation'],
                                    'variance_backlog': std_backlog['variance'],
                                    'index': f[0]['index']}
                                   )

                    closing_stock.clear()
                    shortage_units.clear()
                    quantity_sold.clear()
                    opening_stock.clear()
                    backlog.clear()


    # allow interface to retrieve this level of analysis and then option to summarise the frame

    return summary

def frame(sim_frame):

    cdef int count_runs
    cdef double  revenue, avg_quantity_sold
    cdef:
        list summary=[], item_list=[], min_closing_stock=[], max_closing_stock=[], variance_closing_stock=[],
        average_backlog=[], total_stockout=[]

    cdef list total_revenue=[]
    cdef list max_quantity_sold=[]
    cdef list min_quantity_sold=[]
    cdef list variance_quantity_sold=[]
    cdef list min_opening_stock=[]
    cdef list average_closing_stock=[]

    cdef:
        list max_opening_stock=[], variance_opening_stock=[], min_backlog=[], max_backlog=[], variance_backlog=[],
        average_opening_stock=[], total_shortage_units =[], average_quantity_sold=[]

    cdef unsigned int min_cls, max_cls ,min_opn, max_opn
    cdef double avg_variance_opn, min_bklg, max_bklg, avg_bklg, avg_variance_bklg, std_bklg, avg_variance_cls, std_cls,\
        avg_cls, std_opn, max_qs, min_qs
    cdef float std_quantity_sold, avg_stockout


    sim_frame.sort(key = itemgetter('sku_id'))

    for key, items in itertools.groupby(sim_frame, itemgetter('sku_id')):
        item_list.append(list(items))

    for item in item_list:
        sku_id = item[0]['sku_id']
        count_runs = len(item)


        for j in item:
            total_stockout.append(j['stockout_percentage'])
            average_quantity_sold.append(j['average_quantity_sold'])
            total_shortage_units.append(j['total_shortage_units'])
            min_closing_stock.append(j['minimum_closing_stock'])
            max_closing_stock.append(j['maximum_closing_stock'])
            variance_closing_stock.append(j['variance_closing_stock'])
            min_opening_stock.append(j['minimum_opening_stock'])
            max_opening_stock.append(j['maximum_opening_stock'])
            variance_opening_stock.append(j['variance_opening_stock'])
            min_backlog.append(j['minimum_backlog'])
            max_backlog.append(j['maximum_backlog'])
            variance_backlog.append(j['variance_backlog'])
            max_quantity_sold.append(j['maximum_quantity_sold'])
            min_quantity_sold.append(j['minimum_quantity_sold'])
            variance_quantity_sold.append(j['variance_quantity_sold'])
            average_closing_stock.append(j['average_closing_stock'])
            average_backlog.append(j['average_backlog'])



        min_cls = min(min_closing_stock)
        max_cls = max(max_closing_stock)
        avg_cls = average_items(average_closing_stock, count_runs)
        avg_variance_cls =  average_items(variance_closing_stock, count_runs)
        std_cls = avg_variance_cls ** 0.5
        min_opn = min(min_opening_stock)
        max_opn = max(max_opening_stock)
        avg_variance_opn = average_items(variance_opening_stock, count_runs)
        std_opn = avg_variance_opn ** 0.5
        max_qs = max(max_quantity_sold)
        min_qs = min(min_quantity_sold)
        min_bklg = min(min_backlog)
        max_bklg = max(max_backlog)
        avg_variance_bklg = average_items(variance_backlog, count_runs)
        std_bklg = avg_variance_bklg ** 0.5
        avg_bklg = average_items(average_backlog, count_runs)
        avg_variance = average_items(variance_quantity_sold, count_runs)
        avg_quantity_sold = average_items(average_quantity_sold, count_runs)
        std_quantity_sold = avg_variance ** 0.5
        avg_stockout = average_items(total_stockout, count_runs)
        average_shortage_units = average_items(total_shortage_units, count_runs)
        summary.append({'sku_id': sku_id,
                        'minimum_closing_stock': min_cls,
                        'maximum_closing_stock': max_cls,
                        'average_closing_stock': "{:.0f}".format(avg_cls),
                        'standard_deviation_closing_stock': "{:.0f}".format(std_cls),
                        'service_level': '{:0.2f}'.format((1-avg_stockout)*100),
                        'average_quantity_sold': "{:.0f}".format(avg_quantity_sold),
                        'maximum_quantity_sold': max_qs,
                        'minimum_quantity_sold': min_qs,
                        'minimum_backlog': min_bklg,
                        'maximum_backlog': max_bklg,
                        'average_backlog': "{:.0f}".format(avg_bklg),
                        'standard_deviation_backlog': "{:.0f}".format(std_bklg),
                        'minimum_opening_stock': min_opn,
                        'maximum_opening_stock': max_opn,
                        'variance_opening_stock': "{:.0f}".format(std_opn),
                        'standard_deviation_quantity_sold':"{:.0f}".format(std_quantity_sold),
                        'average_shortage_units': "{:.0f}".format(average_shortage_units)})

        min_closing_stock.clear()
        max_closing_stock.clear()
        min_backlog.clear()
        max_backlog.clear()
        average_backlog.clear()
        min_opening_stock.clear()
        max_opening_stock.clear()
        average_opening_stock.clear()
        max_quantity_sold.clear()
        min_quantity_sold.clear()
        variance_quantity_sold.clear()
        total_revenue.clear()
        variance_closing_stock.clear()
        variance_opening_stock.clear()
        variance_backlog.clear()
        variance_quantity_sold.clear()
        average_closing_stock.clear()
        average_quantity_sold.clear()
        total_shortage_units.clear()
        total_stockout.clear()


        average_shortage_cost = 0
        revenue =0
        avg_stockout = 0

    return summary


def optimise_sim(list orders_analysis , list frame_summary, float service_level):

    cdef float f
    return f


def optimise_service_level(list frame_summary, list orders_analysis, double service_level, int runs,
                           double percentage_increase):
    """ Optimises the safety stock for declared service level.

    Identifies which skus under performed (experiencing a service level lower than expected) after simulating
    transactions over a specific period. The safety stock for these items are increased and the analysis is monte carlo
    is run again.


    Args:
        frame_summary (list):           window summary for each period multiplied by the number of runs.
        orders_analysis (list):         prior analysis of orders data.
        service_level (double):           required service level as a percentage.
        runs (int):                     number of runs from previous
        percentage_increase (double):    the percentage increase required

    Returns:
       list:    Updated orders analysis with new saftey stock values based optimised from the simulation. The initial values
                from the analytical model.

    """
    # compare service levels, build list of skus below service level, change their safety stock increase by a percentage
    # run the monte carlo, keep doing until all skus' are above the requested service level
    # sim_optimise = optimise_sim(service_level=service_level, frame_summary=frame_summary, orders_analysis=orders_analysis)

    count_skus = True

    while count_skus:
        count_skus = False
        for sku in orders_analysis:
            for item in frame_summary:
                if sku.sku_id == item['sku_id']:
                    if float(item['service_level']) <= service_level:
                        count_skus = True
                        sku.safety_stock = round(float(sku.safety_stock)) * percentage_increase
                        break

#       sim = run_monte_carlo(orders_analysis=orders_analysis, runs=runs, period_length=12)

#       sim_window = summarize_window(simulation_frame=sim, period_length=12)

##       sim_frame = summarise_frame(sim_window)

    return orders_analysis



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

from supplychainpy._helpers._pickle_config import deserialise_config, serialise_config

APP_DIR = os.path.dirname(__file__, )
REL_PATH_GENETIC_ALGORITHM = '../sample_data/population_genome.txt'
REL_PATH_DASH = 'dash.pickle'
REL_PATH_ARCHIVE = '../../_archive/'
REL_PATH_CSV_MANAGEMENT_CONFIG = '../_pickled/csv_management_config.pickle'
REL_PATH_APPLICATION_CONFIG = '../_pickled/application_config.pickle'
REL_PATH_PICKLE = '../_pickled/'

ABS_FILE_PATH_DASH = os.path.abspath(os.path.join(APP_DIR, '../_pickled/', REL_PATH_DASH))
ABS_FILE_PATH_APPLICATION_CONFIG = os.path.abspath(os.path.join(APP_DIR, '../_pickled/', REL_PATH_APPLICATION_CONFIG))
ABS_FILE_PATH_CSV_MANAGEMENT_CONFIG = os.path.abspath(os.path.join(APP_DIR, REL_PATH_CSV_MANAGEMENT_CONFIG))
ABS_FILE_PATH_ARCHIVE = os.path.abspath(os.path.join(APP_DIR, REL_PATH_ARCHIVE))
ABS_FILE_GENETIC_ALGORITHM = os.path.abspath(os.path.join(APP_DIR, REL_PATH_ARCHIVE))
ABS_FILE_PICKLE = os.path.abspath(os.path.join(APP_DIR, REL_PATH_PICKLE))

def main():
    print(ABS_FILE_PICKLE)
    #config = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
    #config['file']= 'data4.csv'
    #serialise_config(config, ABS_FILE_PATH_APPLICATION_CONFIG)
    #print(deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG))


if __name__ == '__main__':
    main()



class Currency:
    def __init__(self, code: str):
        self.__code = code

    def __repr__(self):
        return self.retrieve_symbol()

    @property
    def code(self):
        return self.__code

    def retrieve_symbol(self):
        return self._currency_codes(self.code)

    @staticmethod
    def _currency_codes(code: str = None):
        codes = {"AED": u"\u062f"+ u"\u002e" + u"\u0625", "AFN": u"\u060B", "ALL": u"\u004C" + u"\u0065" + u"\u006B",
                 "AMD": u"\u058F", "ANG": "Netherlands Antilles Guilder", "AOA": "Angola Kwanza",
                 "ARS": "Argentina Peso", "AUD": "Australia Dollar", "AWG": "Aruba Guilder",
                 "AZN": "Azerbaijan New Manat", "BAM": "Bosnia and Herzegovina Convertible Marka",
                 "BBD": "Barbados Dollar", "BDT": "Bangladesh Taka", "BGN": "Bulgaria Lev", "BHD": "Bahrain Dinar",
                 "BIF": "Burundi Franc", "BMD": "Bermuda Dollar", "BND": "Brunei Darussalam Dollar",
                 "BOB": "Bolivia Bolíviano", "BRL": "Brazil Real", "BSD": "Bahamas Dollar",
                 "BTN": "Bhutan Ngultrum", "BWP": "Botswana Pula", "BYR": "Belarus Ruble", "BZD": "Belize Dollar",
                 "CAD": "Canada Dollar", "CDF": "Congo/Kinshasa Franc", "CHF": "Switzerland Franc",
                 "CLP": "Chile Peso", "CNY": "China Yuan Renminbi", "COP": "Colombia Peso",
                 "CRC": "Costa Rica Colon", "CUC": "Cuba Convertible Peso", "CUP": "Cuba Peso",
                 "CVE": "Cape Verde Escudo", "CZK": "Czech Republic Koruna", "DJF": "Djibouti Franc",
                 "DKK": "Denmark Krone", "DOP": "Dominican Republic Peso", "DZD": "Algeria Dinar",
                 "EGP": "Egypt Pound", "ERN": "Eritrea Nakfa", "ETB": "Ethiopia Birr",
                 "EUR": u"\u20ac", "FJD": "Fiji Dollar", "FKP": "Falkland Islands (Malvinas) Pound",
                 "GBP": u"\u00A3", "GEL": "Georgia Lari", "GGP": "Guernsey Pound",
                 "GHS": "Ghana Cedi", "GIP": "Gibraltar Pound", "GMD": "Gambia Dalasi", "GNF": "Guinea Franc",
                 "GTQ": "Guatemala Quetzal", "GYD": "Guyana Dollar", "HKD": "Hong Kong Dollar",
                 "HNL": "Honduras Lempira", "HRK": "Croatia Kuna", "HTG": "Haiti Gourde", "HUF": "Hungary Forint",
                 "IDR": "Indonesia Rupiah", "ILS": "Israel Shekel", "IMP": "Isle of Man Pound",
                 "INR": "India Rupee", "IQD": "Iraq Dinar", "IRR": "Iran Rial", "ISK": "Iceland Krona",
                 "JEP": "Jersey Pound", "JMD": "Jamaica Dollar", "JOD": "Jordan Dinar", "JPY": "Japan Yen",
                 "KES": "Kenya Shilling", "KGS": "Kyrgyzstan Som", "KHR": "Cambodia Riel", "KMF": "Comoros Franc",
                 "KPW": "Korea (North) Won", "KRW": "Korea (South) Won", "KWD": "Kuwait Dinar",
                 "KYD": "Cayman Islands Dollar", "KZT": "Kazakhstan Tenge", "LAK:": "Laos Kip",
                 "LBP": "Lebanon Pound", "LKR": "Sri Lanka Rupee", "LRD": "Liberia Dollar", "LSL": "Lesotho Loti",
                 "LYD": "Libya Dinar", "MAD": "Morocco Dirham", "MDL": "Moldova Leu", "MGA": "Madagascar Ariary",
                 "MKD": "Macedonia Denar", "MMK": "Myanmar (Burma) Kyat", "MNT": "Mongolia Tughrik",
                 "MOP": "Macau Pataca", "MRO": "Mauritania Ouguiya", "MUR": "Mauritius Rupee",
                 "MVR": "Maldives (Maldive Islands) Rufiyaa", "MWK": "Malawi Kwacha", "MXN": "Mexico Peso",
                 "MYR": "Malaysia Ringgit", "MZN": "Mozambique Metical", "NAD": "Namibia Dollar",
                 "NGN": "Nigeria Naira", "NIO": "Nicaragua Cordoba", "NOK:": "Norway Krone", "NPR": "Nepal Rupee",
                 "NZD": "New Zealand Dollar", "OMR": "Oman Rial", "PAB": "Panama Balboa", "PEN": "Peru Sol",
                 "PGK": "Papua New Guinea Kina", "PHP": "Philippines Peso", "PKR": "Pakistan Rupee",
                 "PLN": "Poland Zloty", "PYG": "Paraguay Guarani", "QAR": "Qatar Riyal", "RON": "Romania New Leu",
                 "RSD": "Serbia Dinar", "RUB": "Russia Ruble", "RWF": "Rwanda Franc", "SAR": "Saudi Arabia Riyal",
                 "SBD": "Solomon Islands Dollar", "SCR": "Seychelles Rupee", "SDG": "Sudan Pound",
                 "SEK": "Sweden Krona", "SGD": "Singapore Dollar", "SHP": "Saint Helena Pound",
                 "SLL": "Sierra Leone Leone", "SOS": "Somalia Shilling", "SPL:": "Seborga Luigino",
                 "SR:D": "Suriname Dollar", "STD": "São Tomé and Príncipe Dobra", "SVC": "El Salvador Colon",
                 "SYP": "Syria Pound", "SZL": "Swaziland Lilangeni", "THB": "Thailand Baht",
                 "TJS": "Tajikistan Somoni", "TMT": "Turkmenistan Manat", "TND": "Tunisia Dinar",
                 "TOP": "Tonga Pa'anga", "TRY": "Turkey Lira", "TTD": "Trinidad and Tobago Dollar",
                 "TVD": "Tuvalu Dollar", "TWD": "Taiwan New Dollar", "TZS": "Tanzania Shilling",
                 "UAH": "Ukraine Hryvnia", "UGX": "Uganda Shilling", "USD": u"\u0024",
                 "UYU": "Uruguay Peso", "UZS": "Uzbekistan Som", "VEF": "Venezuela Bolivar",
                 "VND": "Viet Nam Dong", "VUV": "Vanuatu Vatu", "WST": "Samoa Tala",
                 "XAF": "Communauté Financière Africaine (BEAC) CFA Franc BEAC", "XCD": "East Caribbean Dollar",
                 "XDR": "International Monetary Fund (IMF) Special Drawing Rights",
                 "XOF": "Communauté Financière Africaine (BCEAO) Franc",
                 "XPF": "Comptoirs Français du Pacifique (CFP) Franc", "YER": "Yemen Rial",
                 "ZAR": "South Africa Rand", "ZMW": "Zambia Kwacha", "ZWD": "Zimbabwe Dollar"}

        symbol = codes.get(code)

        return symbol



import os

import flask
import jinja2
from flask import Flask

from supplychainpy.reporting.blueprints.contact.views import contact_blueprint
from supplychainpy.reporting.blueprints.dashboard.views import dashboard_blueprint, DASHBOARD_TEMPLATE_DIR
from supplychainpy.reporting.blueprints.bot.views import bot_blueprint
from supplychainpy.reporting.blueprints.rawdata.views import rawdata_blueprint
from supplychainpy.reporting.blueprints.recommendations.views import recommendations_blueprint
from supplychainpy.reporting.blueprints.simulation.views import simulation_blueprint

from supplychainpy.reporting.config.settings import ProdConfig, DevConfig
from supplychainpy.reporting.extensions import db, manager
from flask import abort


def create_app(settings_override=None):
    """
    Create flask application using the application factory pattern.

    Returns:
        Flask: Application

    """
    #here = os.path.dirname(os.path.abspath(__file__))
   # static_path = os.path.join(here,'static')
    #print(static_path)

    app = Flask(__name__, instance_relative_config=True, static_folder='static',)

    app.config.from_object(ProdConfig)

    @app.errorhandler(404)
    def page_not_found(e):
        return flask.render_template('404.html'), 404

    if settings_override:
        app.config.update(settings_override)
    try:
        with app.app_context():
            app.register_blueprint(dashboard_blueprint)
            app.register_blueprint(bot_blueprint)
            app.register_blueprint(rawdata_blueprint)
            app.register_blueprint(recommendations_blueprint)
            app.register_blueprint(contact_blueprint)
            app.register_blueprint(simulation_blueprint)
            extensions(app)
    except OSError:
        abort(404)

    return app



def extensions(app):
    """
    Register 0 or more extensions (mutates the app passed in).


    """
    db.init_app(app)
    manager.init_app(app)
    return None



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import os
import pickle

import sys

import logging

from supplychainpy._csv_management._csv_manager import _Orchestrate
from supplychainpy._csv_management._model._db_setup import create_management_db
from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_APPLICATION_CONFIG
from supplychainpy._helpers._decorators import quiet_console
from supplychainpy._helpers._pickle_config import serialise_config, deserialise_config
from supplychainpy.bot.dash import ChatBot
from supplychainpy.launch_reports import launch_load_report, launch_report_server
from supplychainpy.launch_reports import launch_report
from supplychainpy.launch_reports import load_db

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

def main():
    parser = argparse.ArgumentParser(description='Supplychainpy commandline interface a')

    parser.add_argument(dest='filenames', metavar='filename', nargs='?')

    parser.add_argument('-l', '--launch', dest='launch', action='store_true',
                        help='Launches supplychainpy reporting gui for setting \
                        port and launching the default browser.')

    parser.add_argument('-lx', '--launch-console', dest='launch_console', action='store_true',
                        help='Launches supplychainpy reporting in the default browser, \
                        without gui interface. Uses default port (5000) unless another port \
                        is specified. ')

    parser.add_argument('-a', '--analyse', dest='analyse_file', action='store_true',
                        help='Processes the file supplied as the first argument \
                        and performs analysis')

    parser.add_argument('-o', dest='outfile', action='store',
                        help='output file')

    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='verbose mode')

    parser.add_argument('-db', dest='database', action='store',
                        help='database engine uri e.g. ')

    parser.add_argument('-cur', dest='currency', action='store',
                        help='Sets the currency for the analysis. The currency should \
                        match the currency of the raw data. IMPORTANT: Currency \
                        conversion does not occur by setting this flag. The default \
                        currency is US Dollars (USD). ', default='USD')

    parser.add_argument('--host', dest='host', action='store',
                        help='Sets the host for the server \
                        (defaults 127.0.0.1) ', default='127.0.0.1')

    parser.add_argument('--debug', dest='debug', action='store_true',
                        help='Runs in debug mode (default : debug.INFO)')

    parser.add_argument('-loc', dest='location', action='store',
                        help='database path e.g. ')

    parser.add_argument('-u', dest='user', action='store',
                        help='user name ')

    parser.add_argument('-p', '--port', dest='port', action='store',
                        help='port to use for local server e.g. 8080 \
                        (default: 5000)', default='5000')

    parser.add_argument('-c', dest='chat', action='store_true',
                        help='chat to dash from the command line')

    args = parser.parse_args()

    if args.verbose:
        print('filenames = {}'.format(args.filenames))
        print('launch reports = {}'.format(args.launch))
        print('process file = {}'.format(args.analyse_file))
        print('database location = {}'.format(args.location))

    if args.debug:
        logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')


    if args.launch and args.analyse_file is None and args.filenames is not None and args.location:
        # -l -loc
        print(1)

        if args.currency is not None:
            currency = args.currency
        else:
            currency = 'USD'
        app_settings = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
        app_settings['database_path'] = args.location
        app_settings['file'] = args.filenames
        app_settings['currency'] = currency

        serialise_config(app_settings, ABS_FILE_PATH_APPLICATION_CONFIG)
        #d = _Orchestrate()
        #d.copy_file()
        #db_present = d.check_for_db()
        #if db_present:
        #    create_management_db()
        launch_load_report(args.filenames, args.location)

    elif args.launch and args.analyse_file and args.filenames is not None and args.location is not None:
        # -a -loc -l
        print(2)

        if args.currency is not None:
            currency = args.currency
        else:
            currency = 'USD'
        app_settings = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
        app_settings['database_path'] = args.location
        app_settings['file'] = args.filenames
        app_settings['currency'] = currency

        serialise_config(app_settings, ABS_FILE_PATH_APPLICATION_CONFIG)
        #d = _Orchestrate()
        #d.copy_file()
        #db_present = d.check_for_db()
        #if db_present:
        #    create_management_db()
        launch_load_report(args.filenames, args.location)

    elif args.launch and args.location is not None and args.host:
        # -l -loc --host
        print(3)

        app_settings = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)

        app_settings['database_path'] = args.location
        app_settings['port'] = args.port
        app_settings['host'] = args.host,
        app_settings['currency'] = args.currency

        serialise_config(app_settings, ABS_FILE_PATH_APPLICATION_CONFIG)

        launch_report(location=args.location, host=args.host, port=args.port)

    elif args.launch and args.analyse_file and args.filenames and args.location:
        # -a -l -loc
        print(4)

        if args.currency is not None:
            currency = args.currency
        else:
            currency = 'USD'

        app_settings = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
        app_settings['database_path'] = args.location
        app_settings['file'] = args.filenames
        app_settings['currency'] = currency

        serialise_config(app_settings, ABS_FILE_PATH_APPLICATION_CONFIG)
        #d = _Orchestrate()
        #d.copy_file()
        #db_present = d.check_for_db()
        #if db_present:
        #    create_management_db()
        launch_load_report(args.filenames, args.location)
        launch_report(location=args.location, host=args.host, port=args.port)

    elif args.analyse_file and args.location is not None and args.filenames is not None and args.launch_console is None:
        print(8)
        # -a
        app_settings = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
        app_settings['database_path'] = args.location
        app_settings['file'] = args.filenames
        load_db(file=args.filenames, location=args.location)

    elif args.analyse_file and args.location and args.filenames and args.launch_console and args.port:
        # -a -loc --lx -p
        try:
            print(5)
            app_settings = {
                'database_path': args.location,
                'file': args.filenames,
                'currency': args.currency,
                'host': args.host
            }
            serialise_config(app_settings, ABS_FILE_PATH_APPLICATION_CONFIG)
            #d = _Orchestrate()
            #d.copy_file()
            #db_present = d.check_for_db()
            #if db_present:
            #    create_management_db()
            load_db(file=args.filenames, location=args.location)
            launch_report_server(location=args.location, port=args.port, host=args.host)
        except OSError as e:
            print(e)

    elif args.location and args.launch_console and args.port and args.host:
        # -loc -lx -p --host
        try:
            print(6)

            app_settings = {
                'database_path': args.location,
                'host': args.host,
                'currency': args.currency,
                'file': args.filenames
            }
            serialise_config(app_settings, ABS_FILE_PATH_APPLICATION_CONFIG)
            launch_report_server(location=args.location, port=args.port, host=args.host)
        except OSError as e:
            print(e)

    elif args.analyse_file and args.location is not None and args.filenames is not None and args.launch_console is None and args.host:
        # -a -loc -lx --host
        print(7)

        app_settings = {
            'database_path': args.location,
            'host': args.host,
            'currency': args.currency
        }
        serialise_config(app_settings, ABS_FILE_PATH_APPLICATION_CONFIG)
        launch_report_server(location=args.location, port=args.port, host=args.host)

    elif args.chat:
        msg = input("Enter message for Dash: ")
        while msg != 'q':
            dash = ChatBot()
            response = dash.chat_machine(message=msg)
            for i in response:
                print('Dash> ', i)
            msg = input("> ")
    else:
        launch_report()

if __name__ == '__main__':
    main()



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from copy import deepcopy
from typing import List

from supplychainpy.bot._dash_engine import DashMachine
from supplychainpy.bot._dash_states import DashStates
from supplychainpy.bot._helpers import unpack_sentence


class ChatBot:
    """ Chat Bot for supplychainpy Reporting."""

    @staticmethod
    def chat_machine(message: str) -> List[str]:
        """ Interact with chat bot my sending a message and waiting for a response.
        Args:
            message (str):    The message for the chat bot.

        Returns:
            list:   The response from the chat bot.

        Examples:
        >>> chat_bot = ChatBot()
        >>> response = chat_bot.chat_machine(message='hello')
        """
        response = DashMachine()
        states = DashStates()
        states.compiled_response.shared_response.clear()
        response.add_state("start", states.initialise_chat)
        response.add_state("pronoun", states.pronoun)
        response.add_state("greeting", states.greeting)
        response.add_state("deconstruction", states.deconstruction)
        response.add_state("random_utterance", states.random_utterance)
        response.add_state("response", response, end_state=1)
        response.set_start("start")
        response.run(message=message)
        resp = deepcopy(list(states.compiled_response.shared_response.values()))
        states.compiled_response.shared_response.clear()
        return resp[0]


if __name__ == '__main__':
    dude = ChatBot()
    print(dude.chat_machine("you are?"))
    #print(unpack_sentence("Which SKU has the smallest reorder level?"))


# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from sqlalchemy.engine import RowProxy
from sqlalchemy.sql import select

from supplychainpy._csv_management._model._db_setup import (csv_files, transaction_type, logged_transaction)
from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_APPLICATION_CONFIG
from supplychainpy._helpers._pickle_config import deserialise_config


def _retrieve_transaction_type(t_type: str, connection) -> RowProxy:
    """ Retrieves Transaction Type

    Args:
        ttype (str):     The transaction type that represents the trasaction being recorded 'archive' or 'compress'.

    Returns:
        RowProxy:    The transaction_type


    """
    transaction = select([transaction_type.c.type_id]).where(transaction_type.c.type == t_type)

    rp = connection.execute(transaction)
    record = rp.first()
    return record


def post_transaction(analysis_id, connection) -> bool:
    """ Inserts transaction record into csv_management database

    Args:
        analysis_id:    The id for the analysis the data file corresponds to as given by the transaction table,
                        when running an analysis
        connection:     Database connection for controller to use.

    Returns:
        bool:           True for completed transaction and False for failes transaction

    """
    try:
        # Inserts new transaction into transaction log
        transc_type = _retrieve_transaction_type(t_type='archive', connection=connection)
        tr = [i for i in transc_type]
        ins = logged_transaction.insert().values(
            type_id=tr[0]
        )

        result = connection.execute(ins)
        config = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)

        # when file has been copied to archive location the file path should be updated.
        ins = csv_files.insert().values(
            file_path=config.get('file_path'),
            analysis_id=analysis_id,
            file_name='Inventory',
            transaction_id=result.inserted_primary_key[0]
        )

        connection.execute(ins)
        return True
    except OSError:
        return False


if __name__ == '__main__':
    # _retrieve_transaction_type('archive')
    post_transaction(1)



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from _decimal import Decimal
from concurrent.futures import ThreadPoolExecutor

import flask
from flask import Blueprint, request, abort

from supplychainpy import simulate
from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_APPLICATION_CONFIG
from supplychainpy._helpers._enum_formats import FileFormats
from supplychainpy._helpers._pickle_config import deserialise_config
from supplychainpy.model_inventory import analyse
from supplychainpy.reporting.blueprints.models import MasterSkuList
from supplychainpy.reporting.blueprints.simulation.controller import store_simulation, select_last_simulation
from supplychainpy.reporting.extensions import db

simulation_blueprint = Blueprint('simulation', __name__, template_folder='templates')


def page_not_found(e):
    print('hello')
    return flask.render_template('404.html'), 404


def run_simulation():
    pass
    

@simulation_blueprint.route('/simulation', methods=['GET','PUT'])
@simulation_blueprint.route('/simulation/<int:runs>', methods=['GET','PUT'])
def simulation(runs:int=None):
    try:
        database_path = ''
        file_name = ''
        sim_results = []
        sim_summary_results = []
        inventory = []

        if runs is not None:
            config = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
            database_path = config['database_path']
            file_name = config['file']
            file_path = database_path.replace(' ','') + '/' +(file_name.replace(' ',''))
            analysed_orders = analyse(file_path=str(file_path), z_value=Decimal(1.28), reorder_cost=Decimal(5000), file_type=FileFormats.csv.name, length=12, currency='USD')
            # run the simulation, populate a database and then retrieve the most current values for the simulation page.
            try:
                with ThreadPoolExecutor(max_workers=1) as executor:
                    sim = executor.submit(simulate.run_monte_carlo, orders_analysis=analysed_orders, runs=runs, period_length=12)
                    sim_results = sim.result()
                    sim_window = executor.submit(simulate.summarize_window, simulation_frame=sim_results, period_length=12)
                    sim_window_result = sim_window.result()
                    sim_summary = executor.submit(simulate.summarise_frame, sim_window_result)
                    sim_summary_results = sim_summary.result()
                    store_simulation(sim_summary_results)
                inventory = db.session.query(MasterSkuList).all()

            except OSError as e:
                print(e)
        elif runs is None and request.method == 'GET':
            try:
                sim_summary_results = select_last_simulation()
                inventory = db.session.query(MasterSkuList).all()
            except AttributeError as e:
                pass

        return flask.render_template('simulation.html', db= database_path, file_name=file_name, sim=sim_summary_results, runs=sim_results, inventory=inventory)
    except OSError:
        abort(404)

if __name__ == '__main__':
    simulation()


# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

APP_DIR = os.path.dirname(__file__, )
COMPLETE_CSV_SM = 'complete_dataset_small.csv'
COMPLETE_CSV_XSM = 'complete_dataset_xsmall.csv'
COMPLETE_CSV_LG = 'complete_dataset_large.csv'
PARTIAL_COL_CSV_SM ='partial_dataset_col_small.csv'
PARTIAL_ROW_TXT_SM = 'partial_dataset_row_small.txt'
PARTIAL_COL_TXT_SM = 'partial_dataset_col_small.txt'
PARTIAL_CSV_SM = 'partial_dataset_small.csv'
FORECAST_PICKLE = '../sample_data/forecast.pickle'
RECOMMENDATION_PICKLE = '../sample_data/recommendation.pickle'
PROFILE_PICKLE = '../sample_data/profile.pickle'

ABS_FILE_PATH = {'COMPLETE_CSV_SM': os.path.abspath(os.path.join(APP_DIR, COMPLETE_CSV_SM)),
                 'PARTIAL_ROW_TXT_SM': os.path.abspath(os.path.join(APP_DIR, PARTIAL_ROW_TXT_SM)),
                 'COMPLETE_CSV_LG': os.path.abspath(os.path.join(APP_DIR, COMPLETE_CSV_LG)),
                 'PARTIAL_COL_TXT_SM': os.path.abspath(os.path.join(APP_DIR, PARTIAL_COL_TXT_SM)),
                 'PARTIAL_COL_CSV_SM': os.path.abspath(os.path.join(APP_DIR, PARTIAL_COL_CSV_SM)),
                 'PARTIAL_CSV_SM': os.path.abspath(os.path.join(APP_DIR, PARTIAL_CSV_SM)),
                 'FORECAST_PICKLE': os.path.abspath(os.path.join(APP_DIR, FORECAST_PICKLE)),
                 'RECOMMENDATION_PICKLE': os.path.abspath(os.path.join(APP_DIR, RECOMMENDATION_PICKLE)),
                 'PROFILE_PICKLE': os.path.abspath(os.path.join(APP_DIR, PROFILE_PICKLE)),
                 'COMPLETE_CSV_XSM': os.path.abspath(os.path.join(APP_DIR, COMPLETE_CSV_XSM)),}


def main():
    print(ABS_FILE_PATH)


if __name__ == '__main__':
    main()



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import threading
import tkinter as tk
import webbrowser
from tkinter import ttk

import logging

from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_APPLICATION_CONFIG
from supplychainpy._helpers._pickle_config import deserialise_config, serialise_config
from supplychainpy.reporting.app import create_app
from supplychainpy.reporting.blueprints.dashboard.views import db

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

class ReportsLauncher(threading.Thread):
    """Launches reporting lauch panel """

    def __init__(self):
        threading.Thread.__init__(self)
        self.message = "launching reports"
        self.name = "reports"
        self.port = 5000
        self.host = '127.0.0.1'
        self.app = create_app()

    def print_message(self):
        """Prints launch message"""
        print(self.message)


    def run(self):
        config = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
        config['port'] = self.port
        config['host'] = self.host
        serialise_config(config, ABS_FILE_PATH_APPLICATION_CONFIG)
        # print(self.port)
        self.app.run(host=self.host, port=int(self.port))


def exit_report():
    """Exits reports"""
    exit()


def launch_browser(event, url: str):
    """Launches web browser"""
    webbrowser.open_new(str(url))


class SupplychainpyReporting:
    """Creates report launcher gui, to launch browser and using flask local server. Allows port
    number to be changed."""

    def __init__(self, master):
        master.title('Supplychainpy')
        master.resizable(False, False)


        self.spawn = ReportsLauncher()
        self.parent = master
        self.hyperlink = ''
        app_dir = os.path.dirname(__file__, )
        rel_path = 'supplychainpy/reporting/static/images/logo.gif'
        abs_file_path = os.path.abspath(os.path.join(app_dir, '..', rel_path))
        logo = tk.PhotoImage(file=abs_file_path)
        self.style = ttk.Style()
        self.style.configure('TCheckbutton', background='black')
        self.style.configure('TLabel', background='black')

        # set supplychainpy logo for report launcher gui
        self.image = tk.Label(master, image=logo)
        self.image.image = logo
        self.image.config(background='black')
        self.image.grid(row=0, column=1, columnspan=2)

        self.instruction_label = tk.Label(master, text='Launch supplychainpy reports')
        self.instruction_label.grid(row=1, column=1, columnspan=6)
        self.instruction_label.config(background='black', foreground='white')

        self.hyperlink_label = tk.Label(master)
        self.hyperlink_label.config(background='black',
                                    foreground='#8dc53e',
                                    text='click to open browser:',
                                    font=('system', 10, 'bold')
                                   )

        self.validation_label = tk.Label(master)
        self.validation_label.config(background='black', foreground='red',
                                     text='Incorrect port! Please enter correct port number',
                                     font=('system', 10, 'bold'))

        self.runtime_validation_label = tk.Label(master)
        reporting_txt = 'The reports are already running @ {}'.format(self.hyperlink)
        self.runtime_validation_label.config(background='black',
                                             foreground='red',
                                             text=reporting_txt,
                                             font=('system', 10, 'bold'))

        self.port_label = tk.Label(master, text='Change port (default :5000):')
        self.port_label.config(background='black', foreground='white', justify=tk.RIGHT)

        self.port_text = tk.Entry(master, width=10)

        self.change_port = tk.BooleanVar()
        self.change_port.set(False)
        if os.name in ['posix', 'mac']:
            # sort out the checkbutton in ttk not tk
            report_style = ttk.Style()
            report_style.configure(
                'Mac.TCheckbutton',
                focus='#000000',
                background='#000000',
                foreground='#FFFFFF'
                )
            self.change_port_checkbutton = ttk.Checkbutton(
                master,
                variable=self.change_port,
                style='Mac.TCheckbutton',
                text='Change default port (default :5000)',
                command=lambda: self.show_port_entry()
                )

        elif os.name == 'nt':
            print('nt')
            self.change_port_checkbutton = tk.Checkbutton(master, variable=self.change_port, activebackground='black',
                                                          activeforeground='white',
                                                          bg='black', fg='white', relief='solid', selectcolor='blue',
                                                          text='Change default port (default :5000)',
                                                          command=lambda: self.show_port_entry())

        self.change_port_checkbutton.config(onvalue=True)
        self.change_port_checkbutton.grid(row=2, column=1, columnspan=2, pady=(0, 10))

        self.hyperlink_text = tk.Label(master)
        self.hyperlink_text.config(background='black', foreground='lightblue', font=('courier', 11, 'underline'))
        self.hyperlink_text.bind("<Button-1>", lambda e, url=str(self.hyperlink): launch_browser(e, url))

        self.launcher_button = tk.Button(master, bg='black', fg='grey', text='Launch Reporting',
                                         command=lambda: self.spawn_reports()).grid(
            row=6, column=1, pady=(5, 10),
            padx=(15, 5))

        tk.Button(master, bg='black', fg='grey', text='Exit Reporting', command=lambda: exit_report()).grid(row=6,
                                                                                                            column=2,
                                                                                                            pady=(
                                                                                                                5, 10),
                                                                                                            padx=(
                                                                                                                5, 15))

    def spawn_reports(self):
        """Checks if port number is specified, then validates port number."""

        # if port specified check port is numeric
        try:

            if self.port_text.get() is not '' and isinstance(int(self.port_text.get()), int) and self.hyperlink == '':
                self.hyperlink = 'http://{}:{}'.format(self.spawn.host, self.port_text.get())
                self.validation_label.grid_forget()
                self.hyperlink_text.config(text=self.hyperlink)
                self.hyperlink_text.bind("<Button-1>", lambda e, url=str(self.hyperlink): launch_browser(e, url))
                self.hyperlink_text.grid(row=4, column=1, columnspan=2)
                self.hyperlink_label.grid(row=3, column=1, columnspan=2)
                self.spawn.daemon = True
                self.spawn.port = self.port_text.get()
                self.spawn.start()
            elif self.port_text.get() is '':
                self.validation_label.grid_forget()
                self.hyperlink = 'http://{}:{}'.format(self.spawn.host, self.spawn.port)
                self.hyperlink_text.config(text=self.hyperlink)
                self.hyperlink_text.bind("<Button-1>", lambda e, url=str(self.hyperlink): launch_browser(e, url))
                self.hyperlink_text.grid(row=4, column=1, columnspan=2)
                self.hyperlink_label.grid(row=3, column=1, columnspan=2)
                self.spawn.daemon = True
                self.spawn.start()
            else:
                self.hyperlink_label.grid_forget()

        except ValueError:
            self.validation_label.grid(row=3, column=1, columnspan=2)
            self.hyperlink_label.grid_forget()
        except RuntimeError:
            self.runtime_validation_label.grid(row=3, column=1, columnspan=2)

    def show_port_entry(self):
        if self.change_port.get():
            self.port_label.grid(row=5, column=1, columnspan=1, padx=(15, 0), pady=(10, 10))
            self.port_text.grid(row=5, column=2, columnspan=1, padx=(0, 15), pady=(10, 10))
        else:
            self.port_text.forget()
            self.port_label.forget()


def launch_load_report(file: str, location: str = None):
    from supplychainpy.reporting import load
    launcher = tk.Tk()
    app_launch = SupplychainpyReporting(launcher)
    if location is not None:
        load.load(file, location)
    else:
        load.load(file)
    app_launch.parent.configure(background='black')
    launcher.mainloop()


def launch_report(location: str = None, port: int = 5000, host: str = '127.0.0.1'):
    launcher = tk.Tk()

    #pass settings to create_app(settings=)
    app_launch = SupplychainpyReporting(launcher)
    app_launch.spawn.host = host
    app_launch.spawn.port = port
    app_launch.parent.configure(background='black')
    launcher.mainloop()

def load_db(file: str, location: str = None):
    from supplychainpy.reporting import load

    app = create_app()

    if location is not None and os.name in ['posix', 'mac']:
        app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///{}/reporting.db'.format(location)

    elif location is not None and os.name == 'nt':
        app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///{}\\reporting.db'.format(location)

    with app.app_context():
        db.create_all()
    if location is not None:
        load.load(file, location)
    else:
        load.load(file)


def launch_report_server(location: str = None, port: int = 5000, host: str = '127.0.0.1'):
    app = create_app()
    if location is not None and os.name in ['posix', 'mac']:
        app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///{}/reporting.db'.format(location)
    elif location is not None and os.name == 'nt':
        app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///{}\\reporting.db'.format(location)
    with app.app_context():
        db.create_all()
    app_launch = ReportsLauncher()
    app_launch.port = port
    app_launch.host = host
    app_launch.run()





# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

class Costing:
    def __init__(self):
        self.name ="base"

    def print_me(self):
        return print(self.name)


class FIFO(Costing):
    def __init__(self):
        super().__init__()

    def print_me(self):
        return print("FIFO")


class LIFO(Costing):
    def __init__(self):
        super().__init__()


class MovingAverage(Costing):
    def __init__(self):
        super().__init__()
        pass



import os
from Cython.Build import cythonize
from setuptools import setup, find_packages, Extension

extensions = [
    Extension('supplychainpy.simulations.sim_summary',
              ['supplychainpy/simulations/sim_summary.pyx']),
    Extension('supplychainpy.simulations.monte_carlo_simulation',
              ['supplychainpy/simulations/monte_carlo_simulation.pyx']),
    Extension('supplychainpy.inventory.eoq', ['supplychainpy/inventory/eoq.pyx']),
    Extension('supplychainpy.demand._squared_error',
              ['supplychainpy/demand/_squared_error.pyx']),
    Extension('supplychainpy.demand._evo_algo',
              ['supplychainpy/demand/_evo_algo.pyx']),
]
# only required for pypi
here = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(here,'LOG.txt')) as f:
    long_description = f.read().strip()

setup(name='supplychainpy',
      version='0.0.5',
      description='A library for supply chain, operations and manufacturing, analysis, modelling and simulation.',
      url='https://github.com/KevinFasusi/supplychainpy',
      download_url='https://github.com/KevinFasusi/supplychainpy/tarball/0.0.5',
      author='Kevin Fasusi',
      author_email='kevin@supplybi.com',
      license='BSD 3',
      packages=find_packages(exclude=['docs', 'tests', 'scratch.py']),
      test_suite='supplychainpy/tests',
      install_requires=['numpy',
                        'cython',
                        'flask',
                        'scipy',
                        'pandas',
                        'sqlalchemy',
                        'flask-restful',
                        'flask-restless',
                        'flask-script',
                        'flask-sqlalchemy',
                        'flask-uploads',
                        'flask-wtf',
                        'textblob',
                        'Simplejson',
                        'wtforms'
                        ],
      keywords=['supply chain', 'operations research', 'operations management', 'simulation'],
      ext_modules=cythonize(extensions),
      entry_points={
          'console_scripts': [
              'supplychainpy = supplychainpy.supplychain:main'
          ]
      },
      long_description= long_description,
      classifiers=[
        "Development Status :: 1 - Planning",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.5",
    ],
      package_data={
          'supplychainpy': ['reporting/dashboard/templates/dashboard/*','reporting/*','reporting/static/images/*','reporting/static/fonts/*','reporting/static/scripts/*','reporting/static/styles/*','reporting/static/*', 'reporting/templates/*', 'sample_data/*.csv', 'sample_data/*.py','_pickled/*']
      })



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from decimal import Decimal
from operator import attrgetter
from supplychainpy.inventory.abc_xyz import AbcXyz

from supplychainpy.inventory.analyse_uncertain_demand import UncertainDemand


class Inventory:
    def __init__(self, processed_orders):
        self.__analysed_orders = processed_orders
        self.__abc_xyz = self._abc_xyz_summary_raw()

    @property
    def abc_xyz_raw(self):
        return self.__abc_xyz

    def rank_summary(self, attribute: str, count: int = 0, reverse: bool = True) -> dict:
        """ Ranks any SKU in the profile by an attribute from the UncertainDemand objects orders_summary method.

        Args:
            attribute (str):    SKU attribute from analysed orders in UncertainDemand object
            count (int):        The running count required e.g. 10 for top 10.
            reverse (bool):     Reverse the rank order e.g. top 10 or bottom 10

        Returns:
            dict:               Ranked summary for attribute specified.

        """
        if count == 0:
            count = len(self.__analysed_orders)

        try:
            for index, sku in enumerate(sorted(self.__analysed_orders, key=attrgetter(attribute), reverse=reverse)):
                if index > count:
                    break
                yield sku.orders_summary()
        except AttributeError as e:
            possible_attributes = "Incorrect attribute provided as key. Please use one of the following:\n" \
                                  "demand_variability\neconomic_order_quantity\naverage_order\nsafety_stock" \
                                  "\nstandard_deviation\nreorder_level\nreorder_quantity\nrevenue\n" \
                                  "economic_order_quantity\neconomic_order_variable_cost\n" \
                                  "ABC_XYZ_Classification\nexcess_stock\nshortages\nquantity_on_hand"
            raise AttributeError(possible_attributes)

        except TypeError as e:
            print("Failed {}".format(e))

    def _abc_xyz_summary_raw(self):
        abc = AbcXyz(self.__analysed_orders)
        abc.classification_summary()
        # return total excess, shortages,
        return abc

    def abc_xyz_summary(self, classification: tuple = ('AX', 'AY', 'AZ', 'BX', 'BY', 'BZ', 'CX', 'CY', 'CZ'),
                        category: tuple = ('excess_stock', 'shortages', 'revenue'), value: str = 'currency') -> dict:
        """Retrieve currency value or units for key metrics by classification. Defaults to currency unless units stated.

        Args:
            classification (tuple):     ABCXYZ inventory classification of sku.
            category (tuple):           Category with which to aggregate the SKU classification.
            value (str):                Defaults to aggregating classification on currency unless units are specified.

        Returns:
            dict:                       Aggregated summary for classification based on category in
            currency value or units.

        """

        # Retrieves a subset of the orders analysis based on the classification (AX, AY, AZ...) held in dict

        style = {'AX': [analysis if self.__abc_xyz.ax is not None else 0 for analysis in self.__abc_xyz.ax],
                 'AY': [analysis if self.__abc_xyz.ay is not None else 0 for analysis in self.__abc_xyz.ay],
                 'AZ': [analysis if self.__abc_xyz.az is not None else 0 for analysis in self.__abc_xyz.az],
                 'BX': [analysis if self.__abc_xyz.bx is not None else 0 for analysis in self.__abc_xyz.bx],
                 'BY': [analysis if self.__abc_xyz.by is not None else 0 for analysis in self.__abc_xyz.by],
                 'BZ': [analysis if self.__abc_xyz.bz is not None else 0 for analysis in self.__abc_xyz.bz],
                 'CX': [analysis if self.__abc_xyz.cx is not None else 0 for analysis in self.__abc_xyz.cx],
                 'CY': [analysis if self.__abc_xyz.cy is not None else 0 for analysis in self.__abc_xyz.cy],
                 'CZ': [analysis if self.__abc_xyz.cz is not None else 0 for analysis in self.__abc_xyz.cz]}
        try:
            # filters the subset based on classification and category requested and return the currency vale of the

            temp_sum = Decimal(0)
            unit_cost = Decimal(0)

            temp_currency_summary = {}

            for id in classification:
                for label in category:
                    if label in ('excess_stock', 'shortages', 'revenue'):
                        summary = style.get(id)
                        for sku in summary:
                            unit_cost = Decimal(sku.get("unit_cost"))
                            t = {**sku}
                            temp_sum += Decimal(t.get(label))
                            t.clear()

                        if label == 'revenue':
                            temp_currency_summary.update({label: float(temp_sum)})

                        elif label != 'revenue' and value == 'currency':
                            temp_currency_summary.update({'{}_cost'.format(label): float(temp_sum * unit_cost)})

                        else:
                            temp_currency_summary.update({'{}_units'.format(label): float(temp_sum)})

                        temp_sum = 0
                    else:
                        raise TypeError

                yield {id: temp_currency_summary}
                #temp_currency_summary.clear()

        except TypeError as e:
            raise TypeError("abc_xyz_summary terminated {}".format(e))

        except AttributeError as e:
            print("Incorrect Category or Attribute empty. {}".format(e))

    def describe_sku(self, *args: str)->dict:
        """ Summarise SKU using descriptive statistics, key performance indicators, financial fundamentals and rankings

        Args:
            *args (str): SKU unique identification

        Returns:
            dict:   Summary of SKU.

        """
        try:
            for arg in args:
                yield self._summarise_sku(arg)

        except TypeError as e:
            raise TypeError(
                "SKU id {} is not valid, please make sure you supply the correct sku id. {}".format(args, e))
    
    def _summarise_sku(self, sku_id: str) -> dict:
        """Summarise SKU using descriptive statistics, key performance indicators and rankings.

        Args:
            sku_id (str):   SKU unique identification.

        Returns:
            dict:           Summary of SKU.

        """
        selection = UncertainDemand

        for sku in self.__analysed_orders:
            if sku.sku_id == sku_id:
                selection = sku
                break

        # If summary is updated also update possible attributes error in rank summary
        summary = {'sku_id': '{}'.format(selection.sku_id),
                   'revenue_rank': '{}'.format(self._rank(sku_id=sku_id, attribute='revenue')),
                   'revenue': '{}'.format(selection.revenue),
                   'retail_price': '{}'.format(selection.retail_price),
                   'gross_profit_margin': '{}'.format(Decimal(selection.retail_price) - Decimal(selection.unit_cost)),
                   'markup_percentage': '{}'.format(
                       (Decimal(selection.retail_price) - selection.unit_cost) / selection.unit_cost),
                   'unit_cost': '{}'.format(selection.unit_cost),
                   'excess_rank': '{}'.format(self._rank(sku_id=sku_id, attribute='excess_stock_cost')),
                   'excess_units': '{}'.format(selection.excess_stock),
                   'excess_cost': '{}'.format(Decimal(selection.excess_stock_cost)),
                   'shortage_rank': self._rank(sku_id=sku_id, attribute='shortage_cost'),
                   'shortage_units': '{}'.format(round(selection.shortages)),
                   'shortage_cost': '{}'.format(selection.shortage_cost),
                   'safety_stock_units': '{}'.format(round(selection.safety_stock)),
                   'safety_stock_cost': '{}'.format(selection.safety_stock_cost),
                   'safety_stock_rank': '{}'.format(self._rank(sku_id=sku_id, attribute='safety_stock_cost')),
                   'classification': '{}'.format(selection.abcxyz_classification),
                   'average_orders': '{}'.format(round(selection.average_orders)),
                   'min_order': '{}'.format(min(map(int, selection.orders))),
                   'max_order': '{}'.format(max(map(int, selection.orders))),
                   'percentage_contribution_revenue': '{}'.format(selection.percentage_revenue),
                   'quantity_on_hand': '{}'.format(selection.quantity_on_hand),
                   'inventory_turns': '{}'.format((Decimal(selection.total_orders) * Decimal(selection.unit_cost)) / (
                       Decimal(selection.quantity_on_hand) * Decimal(selection.unit_cost))),
                   'inventory_traffic_light': '{}'.format(self._quantity_on_hand_alert(selection)),
                   'unit_cost_rank': self._rank(sku_id=sku_id, attribute='unit_cost'),
                   }
        # print(self._rank(sku_id=sku_id, attribute='shortage_cost'))
        return summary

    def _rank(self, sku_id: str, attribute: str) -> int:
        """ Rank SKU attributes against whole inventory profile.

        Args:
            sku_id (str):       SKU unique identification number.
            attribute (str):    Attribute keyword.

        Returns:
            int:    Attribute rank.

        """
        for index, t in enumerate(sorted(self.__analysed_orders, key=attrgetter(attribute), reverse=True)):
            if t.sku_id == sku_id:
                return index + 1

    def _quantity_on_hand_alert(self, selection: UncertainDemand) -> str:
        """ Indicates whether the quantity of units to hand is (quantity_on_hand < reorder_level(Amber)),
        (quantity_on_hand <50% safety_stock(Red)) and (quantity_on_hand < 75% safety stock(white)

        Args:
            selection (UncertainDemand): Analysed sku object.

        Returns:
            str:    Traffic light indicator summarising final inventory position for inventory quantity on hand.

        """

        half_safety_stock = float(selection.safety_stock) * 0.5
        two_thirds_safety_stock = float(selection.safety_stock) * 0.75
        if selection.reorder_level > selection.quantity_on_hand > selection.safety_stock:
            traffic_light = 'amber'
        elif half_safety_stock > selection.quantity_on_hand > two_thirds_safety_stock:
            traffic_light = 'red'
        elif selection.quantity_on_hand < two_thirds_safety_stock:
            traffic_light = 'white'
        else:
            traffic_light = 'green'

        return traffic_light

    def _currency_codes(self, code: str=None):
        codes = {"AED": u"\u062f"+ u"\u002e" + u"\u0625", "AFN": u"\u060B", "ALL": u"\u004C" + u"\u0065" + u"\u006B",
                 "AMD": "Armenia Dram", "ANG": "Netherlands Antilles Guilder", "AOA": "Angola Kwanza",
                 "ARS": "Argentina Peso", "AUD": "Australia Dollar", "AWG": "Aruba Guilder",
                 "AZN": "Azerbaijan New Manat", "BAM": "Bosnia and Herzegovina Convertible Marka",
                 "BBD": "Barbados Dollar", "BDT": "Bangladesh Taka", "BGN": "Bulgaria Lev", "BHD": "Bahrain Dinar",
                 "BIF": "Burundi Franc", "BMD": "Bermuda Dollar", "BND": "Brunei Darussalam Dollar",
                 "BOB": "Bolivia Bolíviano", "BRL": "Brazil Real", "BSD": "Bahamas Dollar",
                 "BTN": "Bhutan Ngultrum", "BWP": "Botswana Pula", "BYR": "Belarus Ruble", "BZD": "Belize Dollar",
                 "CAD": "Canada Dollar", "CDF": "Congo/Kinshasa Franc", "CHF": "Switzerland Franc",
                 "CLP": "Chile Peso", "CNY": "China Yuan Renminbi", "COP": "Colombia Peso",
                 "CRC": "Costa Rica Colon", "CUC": "Cuba Convertible Peso", "CUP": "Cuba Peso",
                 "CVE": "Cape Verde Escudo", "CZK": "Czech Republic Koruna", "DJF": "Djibouti Franc",
                 "DKK": "Denmark Krone", "DOP": "Dominican Republic Peso", "DZD": "Algeria Dinar",
                 "EGP": "Egypt Pound", "ERN": "Eritrea Nakfa", "ETB": "Ethiopia Birr",
                 "EUR": u"\u20ac", "FJD": "Fiji Dollar", "FKP": "Falkland Islands (Malvinas) Pound",
                 "GBP": u"\u00A3", "GEL": "Georgia Lari", "GGP": "Guernsey Pound",
                 "GHS": "Ghana Cedi", "GIP": "Gibraltar Pound", "GMD": "Gambia Dalasi", "GNF": "Guinea Franc",
                 "GTQ": "Guatemala Quetzal", "GYD": "Guyana Dollar", "HKD": "Hong Kong Dollar",
                 "HNL": "Honduras Lempira", "HRK": "Croatia Kuna", "HTG": "Haiti Gourde", "HUF": "Hungary Forint",
                 "IDR": "Indonesia Rupiah", "ILS": "Israel Shekel", "IMP": "Isle of Man Pound",
                 "INR": "India Rupee", "IQD": "Iraq Dinar", "IRR": "Iran Rial", "ISK": "Iceland Krona",
                 "JEP": "Jersey Pound", "JMD": "Jamaica Dollar", "JOD": "Jordan Dinar", "JPY": "Japan Yen",
                 "KES": "Kenya Shilling", "KGS": "Kyrgyzstan Som", "KHR": "Cambodia Riel", "KMF": "Comoros Franc",
                 "KPW": "Korea (North) Won", "KRW": "Korea (South) Won", "KWD": "Kuwait Dinar",
                 "KYD": "Cayman Islands Dollar", "KZT": "Kazakhstan Tenge", "LAK:": "Laos Kip",
                 "LBP": "Lebanon Pound", "LKR": "Sri Lanka Rupee", "LRD": "Liberia Dollar", "LSL": "Lesotho Loti",
                 "LYD": "Libya Dinar", "MAD": "Morocco Dirham", "MDL": "Moldova Leu", "MGA": "Madagascar Ariary",
                 "MKD": "Macedonia Denar", "MMK": "Myanmar (Burma) Kyat", "MNT": "Mongolia Tughrik",
                 "MOP": "Macau Pataca", "MRO": "Mauritania Ouguiya", "MUR": "Mauritius Rupee",
                 "MVR": "Maldives (Maldive Islands) Rufiyaa", "MWK": "Malawi Kwacha", "MXN": "Mexico Peso",
                 "MYR": "Malaysia Ringgit", "MZN": "Mozambique Metical", "NAD": "Namibia Dollar",
                 "NGN": "Nigeria Naira", "NIO": "Nicaragua Cordoba", "NOK:": "Norway Krone", "NPR": "Nepal Rupee",
                 "NZD": "New Zealand Dollar", "OMR": "Oman Rial", "PAB": "Panama Balboa", "PEN": "Peru Sol",
                 "PGK": "Papua New Guinea Kina", "PHP": "Philippines Peso", "PKR": "Pakistan Rupee",
                 "PLN": "Poland Zloty", "PYG": "Paraguay Guarani", "QAR": "Qatar Riyal", "RON": "Romania New Leu",
                 "RSD": "Serbia Dinar", "RUB": "Russia Ruble", "RWF": "Rwanda Franc", "SAR": "Saudi Arabia Riyal",
                 "SBD": "Solomon Islands Dollar", "SCR": "Seychelles Rupee", "SDG": "Sudan Pound",
                 "SEK": "Sweden Krona", "SGD": "Singapore Dollar", "SHP": "Saint Helena Pound",
                 "SLL": "Sierra Leone Leone", "SOS": "Somalia Shilling", "SPL:": "Seborga Luigino",
                 "SR:D": "Suriname Dollar", "STD": "São Tomé and Príncipe Dobra", "SVC": "El Salvador Colon",
                 "SYP": "Syria Pound", "SZL": "Swaziland Lilangeni", "THB": "Thailand Baht",
                 "TJS": "Tajikistan Somoni", "TMT": "Turkmenistan Manat", "TND": "Tunisia Dinar",
                 "TOP": "Tonga Pa'anga", "TRY": "Turkey Lira", "TTD": "Trinidad and Tobago Dollar",
                 "TVD": "Tuvalu Dollar", "TWD": "Taiwan New Dollar", "TZS": "Tanzania Shilling",
                 "UAH": "Ukraine Hryvnia", "UGX": "Uganda Shilling", "USD": u"\u0024",
                 "UYU": "Uruguay Peso", "UZS": "Uzbekistan Som", "VEF": "Venezuela Bolivar",
                 "VND": "Viet Nam Dong", "VUV": "Vanuatu Vatu", "WST": "Samoa Tala",
                 "XAF": "Communauté Financière Africaine (BEAC) CFA Franc BEAC", "XCD": "East Caribbean Dollar",
                 "XDR": "International Monetary Fund (IMF) Special Drawing Rights",
                 "XOF": "Communauté Financière Africaine (BCEAO) Franc",
                 "XPF": "Comptoirs Français du Pacifique (CFP) Franc", "YER": "Yemen Rial",
                 "ZAR": "South Africa Rand", "ZMW": "Zambia Kwacha", "ZWD": "Zimbabwe Dollar"}

        currency = codes.get('AED')

        return codes

if __name__ == '__main__':
    print('{}'.format(u"\u20ac"))


# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

from scipy import stats

from supplychainpy._helpers._decorators import log_this

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class LinearRegression:
    """ Linear Regression Statistics
    """
    def __init__(self, orders: list):
        self.orders = orders

    @log_this(logging.DEBUG, message='Called to generate Linear regression.')
    def least_squared_error(self, slice_end:int = 0, slice_start: int=0):
        """Calculate simple linear regression values and two_tail pvalue to determine linearity.

        Args:
            slice_end:      Start value for slicing the orders list. Default value is zero.
            slice_start:    End value for slicing the orders list. Default value is Zero. The length of the orders list
                            as supplied to the constructor is then used as the length of the orders.

        Returns:
            dict:           Regression results.

        Examples:
        >>> from supplychainpy.demand._forecast_demand import Forecast
        >>> orders = [165, 171, 147, 143, 164, 160, 152, 150, 159, 169, 173, 203, 169, 166, 162, 147,
        ...           188, 161, 162, 169, 185, 188, 200, 229, 189, 218, 185, 199, 210, 193, 211, 208, 216, 218,
        ...           264, 304]
        >>> total_orders = 0
        >>> for order in orders[:12]:
        >>>     total_orders += order
        >>> avg_orders = total_orders / 12
        >>> forecasting_demand = Forecast(orders, avg_orders)
        >>> forecast = [i for i in forecasting_demand.simple_exponential_smoothing(0.5)]
        >>> regression = LinearRegression(forecast)
        >>> regression_statistics = regression.least_squared_error()
        """

        pvalue = 0.00
        slope = 0.00
        test_statistic = 0.00
        intercept = 0.00
        slope_standard_error = 0.00
        std_residuals = 0.00
        trend = False

        try:
            if slice_end == 0 and slice_start == 0:
                slice_start, slice_end = 0, (len(self.orders) - 1)
            else:
                slice_start = slice_start
                slice_end = slice_end

            x_y_values = {}

            for i in self.orders[slice_start:slice_end]:
                x_y_values.update({i['t']: i['demand']})
            x_mul_y = {k: k * v for k, v in x_y_values.items()}
            x_sq = {k: k ** 2 for k in x_y_values.keys()}
            sum_x = sum([k for k in x_y_values.keys()])
            sum_y = sum([v for v in x_y_values.values()])
            sum_x_sq = sum([x for x in x_sq.values()])
            sum_x_mul_y = sum([v for v in x_mul_y.values()])
            total_sum_x_sq = sum_x ** 2
            total_sum_y_sq = sum_y ** 2
            n = len(x_y_values)
            slope = (n * (sum_x_mul_y) - (sum_x) * (sum_y)) / (n * (sum_x_sq) - total_sum_x_sq)
            intercept = (sum_y / n) - slope * (sum_x / n)
            std_residuals = ((total_sum_y_sq - intercept * sum_y - slope * sum_x_mul_y) / n - 2) ** 0.5
            ss = (sum_x_sq - (total_sum_x_sq) / n)
            slope_standard_error = std_residuals / ss
            test_statistic = slope / slope_standard_error
            pvalue = stats.t.sf(abs(test_statistic), n - 2) * 2

            if pvalue < 0.05:
                trend = True
            else:
                trend = False

        except ValueError as e:
            log.debug('The value supplied to slice the orders list is out of range. {}'.format(e))
            print('Please review the range of orders requested. The current length of the orders is {}. '
                  'Please supply a range within this length.'.format(len(self.orders)))

        return {'slope': slope, 'pvalue': pvalue, 'test_statistic': test_statistic,
                'slope_standard_error': slope_standard_error, 'intercept': intercept,
                'std_residuals': std_residuals, 'trend': trend}
